from v0.relationalai.semantics.metamodel.types import digits_to_bits
from v0.relationalai.semantics.lqp import ir as lqp
from v0.relationalai.semantics.lqp.types import is_numeric
from v0.relationalai.semantics.lqp.utils import TranslationCtx, gen_unique_var, lqp_hash
from v0.relationalai.semantics.lqp.constructors import mk_primitive, mk_specialized_value, mk_type, mk_value

rel_to_lqp = {
    "=": "rel_primitive_eq",
    "!=": "rel_primitive_neq",
    "%": "rel_primitive_remainder",
    "abs": "rel_primitive_abs",
    "ceil": "rel_primitive_round_up",
    "floor": "rel_primitive_round_down",
    "pow": "rel_primitive_power",
    "cbrt": "rel_primitive_cbrt",
    "factorial": "rel_primitive_factorial",
    "cos": "rel_primitive_cos",
    "cosh": "rel_primitive_cosh",
    "acos": "rel_primitive_acos",
    "acosh": "rel_primitive_acosh",
    "sin": "rel_primitive_sin",
    "sinh": "rel_primitive_sinh",
    "asin": "rel_primitive_asin",
    "asinh": "rel_primitive_asinh",
    "tan": "rel_primitive_tan",
    "tanh": "rel_primitive_tanh",
    "atan": "rel_primitive_atan",
    "atanh": "rel_primitive_atanh",
    "cot": "rel_primitive_cot",
    "acot": "rel_primitive_acot",
    "construct_date": "rel_primitive_construct_date",
    "construct_date_from_datetime": "rel_primitive_datetime_date_convert",
    "construct_datetime_ms_tz": "rel_primitive_construct_datetime",
    "hash": "rel_primitive_hash_tuple_uint128",
    "uuid_to_string": "rel_primitive_uuid_string",
    "parse_uuid": "rel_primitive_parse_uuid",
    "parse_date": "rel_primitive_parse_date",
    "parse_datetime": "rel_primitive_parse_datetime",
    "parse_decimal": "rel_primitive_parse_decimal",
    "parse_int64": "rel_primitive_parse_int",
    "parse_int128": "rel_primitive_parse_int128",
    "parse_float": "rel_primitive_parse_float",
    "string": "rel_primitive_string",
    "starts_with": "rel_primitive_starts_with",
    "ends_with": "rel_primitive_ends_with",
    "contains": "rel_primitive_contains",
    "num_chars": "rel_primitive_num_chars",
    "substring": "rel_primitive_substring",
    "like_match": "rel_primitive_like_match",
    "lower": "rel_primitive_lowercase",
    "upper": "rel_primitive_uppercase",
    "concat": "rel_primitive_concat",
    "replace": "rel_primitive_replace",
    "strip": "rel_primitive_trim",
    "date_year": "rel_primitive_date_year",
    "date_quarter": "rel_primitive_date_quarterofyear",
    "date_month": "rel_primitive_date_month",
    "date_week": "rel_primitive_date_week",
    "date_day": "rel_primitive_date_day",
    "date_dayofyear": "rel_primitive_date_dayofyear",
    "date_weekday": "rel_primitive_date_dayofweek",
    "date_add": "rel_primitive_typed_add_date_period",
    "date_subtract": "rel_primitive_typed_subtract_date_period",
    "dates_period_days": "rel_primitive_date_days_between",
    "datetime_now": "__pyrel_lqp_intrinsic_datetime_now",
    "datetime_add": "rel_primitive_typed_add_datetime_period",
    "datetime_subtract": "rel_primitive_typed_subtract_datetime_period",
    "datetime_year": "rel_primitive_datetime_year",
    "datetime_quarter": "rel_primitive_datetime_quarterofyear",
    "datetime_month": "rel_primitive_datetime_month",
    "datetime_week": "rel_primitive_datetime_week",
    "datetime_day": "rel_primitive_datetime_day",
    "datetime_dayofyear": "rel_primitive_datetime_dayofyear",
    "datetime_hour": "rel_primitive_datetime_hour",
    "datetime_minute": "rel_primitive_datetime_minute",
    "datetime_second": "rel_primitive_datetime_second",
    "datetime_weekday": "rel_primitive_datetime_dayofweek",
    "datetimes_period_milliseconds": "rel_primitive_datetime_milliseconds_between",
    "date_format": "rel_primitive_format_date",
    "datetime_format": "rel_primitive_format_datetime",
    "range": "rel_primitive_range",
    "natural_log": "rel_primitive_natural_log",
    "log": "rel_primitive_log",
    "log2": "rel_primitive_log2",
    "log10": "rel_primitive_log10",
    "sqrt": "rel_primitive_sqrt",
    "isinf": "rel_primitive_isinf",
    "isnan": "rel_primitive_isnan",
    "exp": "rel_primitive_natural_exp",
    "erf": "rel_primitive_error_function",
    "erfinv": "rel_primitive_error_function_inverse",
    # Division is monotype, but only on the input args. Until we distinguish between input
    # and output args, we can't use the same assertions for monotype-ness as the other ops.
    "/": "rel_primitive_divide_monotype",
    "levenshtein": "rel_primitive_levenshtein",
    "split": "rel_primitive_string_split",
    "split_part": "rel_primitive_string_split",
    "regex_match": "rel_primitive_regex_match",
    "infomap": "rel_primitive_infomap",
    "louvain": "rel_primitive_louvain",
    "label_propagation": "rel_primitive_async_label_propagation",
}

primitive_type_reorderings = {
    "rel_primitive_like_match": [1, 0],
}

agg_to_lqp = {
    "min": "rel_primitive_min",
    "max": "rel_primitive_max",
    "sum": "rel_primitive_add_monotype",
    "count": "rel_primitive_add_monotype", # count is a sum of 1s
    "rel_primitive_solverlib_ho_appl": "rel_primitive_solverlib_ho_appl",
}

rel_to_lqp_monotype = {
    "+": "rel_primitive_add_monotype",
    "-": "rel_primitive_subtract_monotype",
    "*": "rel_primitive_multiply_monotype",
    "<=": "rel_primitive_lt_eq_monotype",
    ">=": "rel_primitive_gt_eq_monotype",
    ">": "rel_primitive_gt_monotype",
    "<": "rel_primitive_lt_monotype",
    "//": "rel_primitive_trunc_divide_monotype",
    "maximum": "rel_primitive_max",
    "minimum": "rel_primitive_min",
}

# Insert extra terms where a raicode primitive expects more terms, and there are possible
# defaults.
def _extend_primitive_terms(name: str, terms: list[lqp.RelTerm], term_types: list[lqp.Type]) -> tuple[list[lqp.RelTerm], list[lqp.Type]]:
    if name == "rel_primitive_parse_decimal" and len(terms) == 2:
        assert term_types
        py_precision = term_types[1].parameters[0].value
        bit_value = mk_value(digits_to_bits(py_precision))
        precision_value = mk_value(term_types[1].parameters[1].value)
        terms = [
            mk_specialized_value(bit_value),
            mk_specialized_value(precision_value),
            terms[0],
            terms[1],
        ]
        term_types = [mk_type(lqp.TypeName.INT), mk_type(lqp.TypeName.INT), *term_types]

    return (terms, term_types)

# Reorder terms where the raicode primitive expects them in a different order.
def _reorder_primitive_terms(name: str, terms: list[lqp.RelTerm], term_types: list[lqp.Type]) -> tuple[list[lqp.RelTerm], list[lqp.Type]]:
    reordering = primitive_type_reorderings.get(name, None)
    if reordering:
        assert len(terms) == len(reordering), \
            f"Primitive {name} expected {len(reordering)} terms, got {len(terms)}"
        return ([terms[i] for i in reordering], [term_types[i] for i in reordering])

    return (terms, term_types)

# Check that the primitive terms have the expected types.
# TODO: Fill out for all primitives, or rely on server-side type checking?
def _assert_primitive_terms(name: str, terms: list[lqp.RelTerm], term_types: list[lqp.Type]):
    if is_monotype(name):
        # Make sure that the input terms have the same types
        assert term_types.count(term_types[0]) == len(term_types), \
            f"Expected all terms to have the same type for monotype operator " \
            f"`{name}` but got {term_types} for terms {terms}"
    elif name == "=":
        # Allowed types are monotype, or both number types
        monotype = term_types[0] == term_types[1]
        numeric = is_numeric(term_types[0]) and is_numeric(term_types[1])
        assert monotype or numeric, \
            f"Expected types of `=`to be comparable but got `{[t.type_name for t in term_types]}`"

def build_primitive(
    name: str,
    terms: list[lqp.RelTerm],
    term_types: list[lqp.Type]
) -> lqp.Formula:
    lqp_name = relname_to_lqp_name(name)
    terms, term_types = _extend_primitive_terms(lqp_name, terms, term_types)
    terms, term_types = _reorder_primitive_terms(lqp_name, terms, term_types)
    _assert_primitive_terms(lqp_name, terms, term_types)

    # Handle intrinsics. To callers of `build_primitive` the distinction between intrinsic
    # and primitive doesn't matter, so we don't want to burden them with that detail.
    # Intrinsics are built-in definitions added by the LQP emitter, that user logic can just
    # refer to.
    if lqp_name == "__pyrel_lqp_intrinsic_datetime_now":
        id = lqp.RelationId(id=lqp_hash(lqp_name), meta=None)
        assert len(terms) == 1
        assert isinstance(terms[0], lqp.Term)
        return lqp.Atom(name=id, terms=[terms[0]], meta=None)

    return mk_primitive(lqp_name, terms)

def relname_to_lqp_name(name: str) -> str:
    if name in rel_to_lqp:
        return rel_to_lqp[name]
    elif name in rel_to_lqp_monotype:
        return rel_to_lqp_monotype[name]
    else:
        # If we don't have a mapping for the built-in, we just pass it through as-is.
        return name

def is_monotype(name: str) -> bool:
    return name in rel_to_lqp_monotype

# We take the name and type of the variable that we're summing over, so that we can generate
# recognizable names for the variables in the reduce operation and preserve the type.
def lqp_avg_op(ctx: TranslationCtx, op_name: str, sum_name: str, sum_type: lqp.Type) -> lqp.Abstraction:
    count_type = mk_type(lqp.TypeName.INT)
    vars = [
        (gen_unique_var(ctx, sum_name), sum_type),
        (gen_unique_var(ctx, "counter"), count_type),
        (gen_unique_var(ctx, sum_name), sum_type),
        (gen_unique_var(ctx, "one"), count_type),
        (gen_unique_var(ctx, "sum"), sum_type),
        (gen_unique_var(ctx, "count"), count_type),
    ]

    x1 = vars[0][0]
    x2 = vars[1][0]
    y1 = vars[2][0]
    y2 = vars[3][0]
    sumv = vars[4][0]
    count = vars[5][0]

    body = lqp.Conjunction(
        args=[
            mk_primitive("rel_primitive_add_monotype", [x1, y1, sumv]),
            mk_primitive("rel_primitive_add_monotype", [x2, y2, count])
        ],
        meta=None
    )
    return lqp.Abstraction(vars=vars, value=body, meta=None)

# Default handler for aggregation operations in LQP.
def lqp_agg_op(ctx: TranslationCtx, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
    x = gen_unique_var(ctx, f"x_{aggr_arg_name}")
    y = gen_unique_var(ctx, f"y_{aggr_arg_name}")
    z = gen_unique_var(ctx, f"z_{aggr_arg_name}")
    ts = [(x, aggr_arg_type), (y, aggr_arg_type), (z, aggr_arg_type)]

    name = agg_to_lqp.get(op_name, op_name)
    body = mk_primitive(name, [x, y, z])

    return lqp.Abstraction(vars=ts, value=body, meta=None)

def lqp_operator(ctx: TranslationCtx, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
    # TODO: Can we just pass through unknown operations?
    if op_name not in agg_to_lqp:
        raise NotImplementedError(f"Unsupported aggregation: {op_name}")

    return lqp_agg_op(ctx, op_name, aggr_arg_name, aggr_arg_type)



import warnings

from v0.relationalai.semantics.metamodel.compiler import Pass, group_tasks

__all__ = ["Pass", "group_tasks"]

warnings.warn(
    "relationalai.early_access.metamodel.compiler is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.compiler",
    DeprecationWarning,
    stacklevel=2,
)


import base64
import hashlib
import json
import jwt
import requests
import secrets
import threading
import urllib.parse
import webbrowser

from abc import ABC, abstractmethod
from datetime import timedelta, timezone, datetime
from typing import Dict, Tuple

from .jwt_generator import JWTGenerator
from .oauth_callback_server import OAuthCallbackServer
from .util import is_keyring_secure, is_localhost, extract_port, remove_url_scheme
from ..debugging import warn
from ..errors import RAIException, DirectAccessInvalidAuthException, InsecureKeychainWarning, KeychainFailureWarning
from ..tools.constants import SNOWFLAKE_DIRECT_ACCESS_AUTHS
from ..clients.config import Config

# keyring is not available in the snowflake notebook environment
try:
    import keyring
    KEYRING_AVAILABLE = True
except ImportError:
    KEYRING_AVAILABLE = False
    keyring = None

class TokenHandler(ABC):
    def __init__(self, config: Config):
        super().__init__()
        self._config = config
        self.role: str = str(self._config.get("role", ""))
        self.user: str = str(self._config.get("user", ""))
        self.account: str = self._prepare_account_name(str(config.get("account", "")))
        self._renewal_buffer_time = timedelta(minutes=2)

    @staticmethod
    def from_config(config: Config) -> "TokenHandler":
        authenticator = config.get("authenticator", "")
        if authenticator == "oauth_authorization_code":
            return SnowflakeOAuthTokenHandler(config)
        if authenticator == "programmatic_access_token":
            return SnowflakePATHandler(config)
        if authenticator == "snowflake_jwt":
            return SnowflakeKeyFileHandler(config)

        raise DirectAccessInvalidAuthException(authenticator, SNOWFLAKE_DIRECT_ACCESS_AUTHS.keys())

    @abstractmethod
    # returns a token which can be used to create a snowflake session object
    def get_session_login_token(self) -> str:
        pass

    @abstractmethod
    # returns a token which can be used to access a specific endpoint in multiple cases this
    # can be the same token as for session login but is dependent on the authentication
    # method
    def get_ingress_token(self, endpoint) -> str:
        pass

    @staticmethod
    # this helper method is needed to prepare the account name to a format accepted to form 
    # the snowflake URLs. Without the replacement of '_' with '-' the requests to snowflake
    # auth endpoints will lead to SSL mismatch errors.
    def _prepare_account_name(account: str) -> str:
        return account.replace("_", "-").upper()

    # helper method to retrieve a token from a cache and check if it is expired
    # assumes the key consisting of the (account, user, role) tuple
    def _get_token_from_cache(self, cache) -> Tuple:
        key = (self.account, self.user, self.role)
        if key in cache:
            token, expiration_timestamp = cache[key]
            current_timestamp = datetime.now(timezone.utc).timestamp()
            # only use cached token if not expired and not close to expiring
            if expiration_timestamp - self._renewal_buffer_time.total_seconds() > current_timestamp:
                return token, False
            else:
                return "", True
        return "", False

class SnowflakeOAuthTokenHandler(TokenHandler):

    # Global cache to re-use the tokens over multiple threads and resource objects
    # Maps (account, user, role) to (token, expiration_timestamp)
    _token_cache: Dict[Tuple[str, str, str], Tuple[str, float]] = dict()
    _refresh_token_cache: Dict[Tuple[str, str, str], Tuple[str, float]] = dict()
    
    # Global cache to re-use auth-tokens over multiple threads and resource objects
    # Maps (account, user, role) to ((auth_code, code_verifier), expiration_time)
    # This additional cache is needed to avoid re-authentication if token exchange fails
    # in one of the threads.
    _auth_token_cache: Dict[Tuple[str, str, str], Tuple[Tuple[str, str], float]] = dict()

    # To avoid authenticating for multiple threads, we use this lock to let the first thread
    # authenticate and the others wait for the result retrieved from the global cache
    _authenticate_lock = threading.Lock()

    SERVICE_NAME = "RAI_PYREL_OAUTH"

    def __init__(self, config: Config):
        super().__init__(config)
        self.client_id: str = config.get("oauth_client_id", "") or ""
        if not self.client_id:
            raise RAIException("OAuth client ID is not configured. Please set 'oauth_client_id' in your configuration.")
        # client secret not mandatory, only needed for non-public security integrations
        self.client_secret: str = config.get("oauth_client_secret", "") or ""
        self.redirect_uri: str = config.get("oauth_redirect_uri", "") or ""
        if not self.redirect_uri:
            raise RAIException("OAuth redirect URI is not configured. Please set 'oauth_redirect_uri' in your configuration.")
        if not is_localhost(self.redirect_uri):
            raise RAIException("OAuth redirect URI must be a localhost URI.")
        self.redirect_port: int = extract_port(self.redirect_uri)
        self.account_url: str = f"https://{self.account}.snowflakecomputing.com"
        self._callback_server = None
        self._load_refresh_token_cache()

    # performing token retrieval in a thread safe manner. This is needed as multiple
    # resources are oftentimes created in parallel, leading to multiple token handlers
    # hitting the Snowflake auth endpoint at the same time and trying to start local
    # callback servers. Instead we re-suse the retrieved token until the end of its
    # lifetime over all threads.
    def get_session_login_token(self) -> str:
        token, expired = self._get_access_token_from_cache()
        if token and not expired:
            return token

        with SnowflakeOAuthTokenHandler._authenticate_lock:
            token, expired = self._get_access_token_from_cache()
            if token and not expired:
                return token
            refresh_token, expired_refresh = self._get_refresh_token_from_cache()
            if refresh_token and not expired_refresh:
                return self._refresh_token()

            auth_code, code_verifier = self._authenticate()
            return self._exchange_code_for_token(auth_code, code_verifier)
        
    def get_ingress_token(self, endpoint) -> str:
        return self.get_session_login_token()

    def _get_access_token_from_cache(self) -> Tuple[str, bool]:
        return self._get_token_from_cache(SnowflakeOAuthTokenHandler._token_cache)

    def _get_refresh_token_from_cache(self) -> Tuple[str, bool]:
        return self._get_token_from_cache(SnowflakeOAuthTokenHandler._refresh_token_cache)

    def _get_auth_token_from_cache(self) -> Tuple[str, str, bool]:
        key = (self.account, self.user, self.role)
        if key in SnowflakeOAuthTokenHandler._auth_token_cache:
            (auth_code, code_verifier), expiration_timestamp = SnowflakeOAuthTokenHandler._auth_token_cache[key]
            current_timestamp = datetime.now(timezone.utc).timestamp()
            if expiration_timestamp > current_timestamp:
                return auth_code, code_verifier, False
            else:
                return "", "", True
        return "", "", False

    def _load_refresh_token_cache(self):
        with SnowflakeOAuthTokenHandler._authenticate_lock:
            if not KEYRING_AVAILABLE or keyring is None:
                return
            try:
                if not is_keyring_secure():
                    warn(InsecureKeychainWarning())
                    return
                key = (self.account, self.user, self.role)
                if key in SnowflakeOAuthTokenHandler._refresh_token_cache:
                    return
                raw = keyring.get_password(SnowflakeOAuthTokenHandler.SERVICE_NAME, json.dumps(key))
                if not raw:
                    return
                SnowflakeOAuthTokenHandler._refresh_token_cache[key] = tuple(json.loads(raw))
            except Exception:
                warn(KeychainFailureWarning("Failed to load oauth refresh token from keyring."))

    def _persist_refresh_token_cache_entry(self, cache_entry: Tuple[str, float]):
        try:
            if not KEYRING_AVAILABLE or keyring is None:
                return
            if not is_keyring_secure():
                warn(InsecureKeychainWarning())
                return
            key = (self.account, self.user, self.role)
            keyring.set_password(
                SnowflakeOAuthTokenHandler.SERVICE_NAME, 
                json.dumps(key), 
                json.dumps(cache_entry),
            )
        except Exception:
            warn(KeychainFailureWarning("Failed to persist oauth refresh token to keyring."))

    @staticmethod
    def _generate_pkce_challenge() -> Tuple[str, str]:
            """Generate PKCE code verifier and challenge for OAuth flow"""
            code_verifier = base64.urlsafe_b64encode(secrets.token_bytes(32)).decode('utf-8').rstrip('=')
            code_challenge = base64.urlsafe_b64encode(
                hashlib.sha256(code_verifier.encode('utf-8')).digest()
            ).decode('utf-8').rstrip('=')
            
            return code_verifier, code_challenge

    def _get_authorization_url(self, state, code_challenge) -> str:
        """
        Build Snowflake OAuth authorization URL
        
        Args:
            state (str): OAuth state parameter
            code_challenge (str): PKCE code challenge
            scopes (list, optional): OAuth scopes
            
        Returns:
            str: Authorization URL
        """

        params = {
            'client_id': self.client_id,
            'redirect_uri': self.redirect_uri,
            'response_type': 'code',
            'state': state,
            'code_challenge': code_challenge,
            'code_challenge_method': 'S256'
        }
        role = self._config.get("role")
        if role:
            params['scope'] = f"session:role:{role}"
        auth_url = f"{self.account_url}/oauth/authorize?" + urllib.parse.urlencode(params)
        return auth_url

    def _authenticate(self) -> Tuple[str, str]:
        """
        Perform complete OAuth authentication flow using a socket-based callback server
        """
        auth_code, code_verifier, expired = self._get_auth_token_from_cache()
        if not expired and auth_code and code_verifier:
            return auth_code, code_verifier
        
        print("Starting Snowflake OAuth authentication...")
        code_verifier, code_challenge = self._generate_pkce_challenge()
        state = secrets.token_urlsafe(32)
        auth_url = self._get_authorization_url(state, code_challenge)
        self._callback_server = OAuthCallbackServer("localhost", self.redirect_port)
        print("Opening browser for authentication...")
        webbrowser.open(auth_url)
        print("Waiting for authentication callback...")
        self._callback_server.wait_for_callback(timeout=300)
        if self._callback_server.auth_error:
            raise RAIException(f"OAuth authentication failed: {self._callback_server.auth_error}")
        if self._callback_server.auth_state != state:
            raise RAIException("OAuth state mismatch - possible security issue")

        expiration_timestamp = (datetime.now(timezone.utc) + timedelta(seconds=60)).timestamp()
        SnowflakeOAuthTokenHandler._auth_token_cache[(self.account, self.user, self.role)] = (
            (self._callback_server.auth_code, code_verifier), 
            expiration_timestamp,
        )

        return self._callback_server.auth_code, code_verifier
        
    def _exchange_code_for_token(self, auth_code, code_verifier) -> str:
        """
        Exchange authorization code for access token
        
        Args:
            auth_code (str): Authorization code from callback
            code_verifier (str): PKCE code verifier
            
        Returns:
            dict: Token response containing access_token, token_type, etc.
        """
        token_url = f"{self.account_url}/oauth/token-request"
        
        issue_time = datetime.now(timezone.utc)
        data = {
            'grant_type': 'authorization_code',
            'code': auth_code,
            'redirect_uri': self.redirect_uri,
            'code_verifier': code_verifier,
            'client_id': self.client_id
        }
        
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        auth = None
        if self.client_secret:
            auth = (self.client_id, self.client_secret)

        try:
            response = requests.post(token_url, data=data, headers=headers, auth=auth)
        except Exception as e:
            raise RAIException(f"Error during token exchange: {e}")
        if response.status_code != 200:
            raise RAIException(f"Token exchange failed: {response.status_code} - {response.text}")

        return self._handle_token_response(response, issue_time)

    def _refresh_token(self) -> str:
        refresh_token, expired = self._get_refresh_token_from_cache()

        if not refresh_token or expired:
            auth_code, code_verifier = self._authenticate()
            return self._exchange_code_for_token(auth_code, code_verifier)

        issue_time = datetime.now(timezone.utc)
        token_url = f"{self.account_url}/oauth/token-request"
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
            'client_id': self.client_id
        }
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        auth = None
        if self.client_secret:
            auth = (self.client_id, self.client_secret)

        try:
            response = requests.post(token_url, data=data, headers=headers, auth=auth)
        except Exception as e:
            raise RAIException(f"Error during token exchange: {e}")
        if response.status_code != 200:
            raise RAIException(f"Token refresh failed: {response.status_code} - {response.text}")

        return self._handle_token_response(response, issue_time, is_refresh=True)

    def _handle_token_response(self, response: requests.Response, issue_time: datetime, is_refresh: bool = False) -> str:
        if response.status_code == 200:
            response_json = response.json()

            access_token = response_json.get("access_token")
            token_expire_seconds = response_json.get("expires_in")
            expiration_timestamp = (issue_time + timedelta(seconds=token_expire_seconds)).timestamp()
            SnowflakeOAuthTokenHandler._token_cache[(self.account, self.user, self.role)] = (access_token, expiration_timestamp)

            if not is_refresh:
                refresh_token = response_json.get("refresh_token")
                refresh_token_expire_seconds = response_json.get("refresh_token_expires_in")
                refresh_expiration_timestamp = (issue_time + timedelta(seconds=refresh_token_expire_seconds)).timestamp()
                cache_entry = (refresh_token, refresh_expiration_timestamp)
                SnowflakeOAuthTokenHandler._refresh_token_cache[(self.account, self.user, self.role)] = cache_entry
                self._persist_refresh_token_cache_entry(cache_entry)

            return access_token
        else:
            raise Exception(f"Failed to retrieve token: {response.status_code} - {response.text}")

class SnowflakePATHandler(TokenHandler):
    def __init__(self, config: Config):
        super().__init__(config)
        self._token = None
        self.token_file_path: str = str(self._config.get("token_file_path", ""))
        if not self.token_file_path:
            raise RAIException("Token file path is not configured. Add 'token_file_path' to your configuration.")

    @property
    def token(self) -> str:
        if not self._token:
            self._token = self._read_token_file()
        return self._token

    def _read_token_file(self) -> str:
        try:
            with open(self.token_file_path, "r") as f:
                return f.read().strip()
        except Exception as e:
            raise RAIException(f"Failed to read personal access token file: {e}")

    def get_session_login_token(self) -> str:
        return self.token

    def get_ingress_token(self, endpoint) -> str:
        return self.token

class SnowflakeKeyFileHandler(TokenHandler):
    # Global cache to re-use the tokens over multiple threads and resource objects
    # Maps (account, user, role) to (token, expiration_timestamp)
    _token_cache: Dict[Tuple[str, str, str], Tuple[str, float]] = dict()

    # To avoid authenticating for multiple threads, we use this lock to let the first thread
    # authenticate and the others wait for the result retrieved from the global cache
    _authenticate_lock = threading.Lock()

    def __init__(self, config: Config):
        super().__init__(config)
        self.key_file_path = self._config.get("private_key_file")
        if not self.key_file_path:
            raise RAIException("Key file path is not configured.")
        self.key_file_pwd = self._config.get("private_key_file_pwd", "")
        self._jwt_generator = JWTGenerator(self.account, self.user, self.key_file_path, self.key_file_pwd)

    def get_session_login_token(self) -> str:
        return self._jwt_generator.get_or_generate_token()
    
    # performing token retrieval in a thread safe manner. This is needed as multiple
    # resources are oftentimes created in parallel, leading to multiple token handlers
    # hitting the Snowflake auth endpoint at the same time. Instead we re-suse the retrieved
    # token until the end of its lifetime over all threads.
    def get_ingress_token(self, endpoint) -> str:
        with SnowflakeKeyFileHandler._authenticate_lock:
            token, expired = self._get_token_from_cache(SnowflakeKeyFileHandler._token_cache)
            if token and not expired:
                return token
            
            generated_jwt_token = self.get_session_login_token()
            token = self._exchange_token(generated_jwt_token, endpoint)
            return token

    def _exchange_token(self, token, endpoint):
        endpoint = remove_url_scheme(endpoint)
        scope_role = f'session:role:{self.role}' if self.role is not None else None
        scope = f'{scope_role} {endpoint}' if scope_role is not None else endpoint
        data = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'scope': scope,
            'assertion': token,
        }
        url = f'https://{self.account}.snowflakecomputing.com/oauth/token'
        try:
            response = requests.post(url, data=data)
        except Exception as e:
            raise RAIException(f"Error during token exchange: {e}")
        if response.status_code != 200:
            raise RAIException(f"Token exchange failed: {response.status_code} - {response.text}")
        token = response.text
        try:
            expire_timestamp = jwt.decode(token, options={"verify_signature": False})['exp']
        except Exception as e:
            raise RAIException(f"Failed to decode exchanged token: {e}")
        SnowflakeKeyFileHandler._token_cache[(self.account, self.user, self.role)] = (token, expire_timestamp)
        return token



import typing
from .rpq.rpq import RPQ, Segment, _process_segment, Star
from .rpq.rpq import Node, Edge, Concat, Optional, Plus, Union
from .rpq.filter import AnonymousNodeFilter
from .find_paths_via_automaton import find_paths_via_automaton
from .product_graph import project_product_graph_paths_to_original_graph

from v0.relationalai.semantics import Concept, Relationship, define

#-----------------------------------------------------------------------------
# User-facing RPQ constructors
#-----------------------------------------------------------------------------

# (filter)
def node_filter(filter: typing.Callable) -> RPQ:
    from .rpq import Node

    return Node(label = None, filter = AnonymousNodeFilter(filter))


# (node_lab {∧ filter})
def node(node_lab: typing.Union[str,Concept],
         filter: typing.Optional[typing.Callable] = None) -> RPQ:
    from .rpq.filter import NodeLabel, parse_label

    filter = AnonymousNodeFilter(filter) if filter is not None else None

    if isinstance(node_lab, Concept):
        label = NodeLabel(node_lab._name)
    else:
        label = parse_label(node_lab)
    assert isinstance(label, NodeLabel), f"Expected a node label, got {label}"

    return Node(label, filter)

# -[edge_lab {∧ filter}]- or <-[edge_lab {∧ filter}]-
def edge(edge_lab:str, filter: typing.Optional[typing.Callable] = None) -> RPQ:
    from .rpq.filter import EdgeLabel, AnonymousEdgeFilter, parse_label

    label = parse_label(edge_lab)
    assert isinstance(label, EdgeLabel), f"Expected an edge label, either `-[...]->` or `<-[...]-`, got {label}"
    filter = AnonymousEdgeFilter(filter) if filter is not None else None
    return Edge(label, filter)


# (expr₁ ⋅ ... ⋅ exprₙ)
def path(segment: Segment, *segments: Segment) -> RPQ:

    pattern = _process_segment(segment)
    for seg in segments:
        pattern = Concat(pattern, _process_segment(seg))

    return pattern

# (expr₁ ⋅ ... ⋅ exprₙ)?
def optional(*segments: Segment) -> RPQ:

    return Optional(path(*segments))

# (expr₁ ⋅ ... ⋅ exprₙ)+
def plus(segment: Segment, *segments: Segment) -> RPQ:

    return Plus(path(segment, *segments))

# (expr₁ ⋅ ... ⋅ exprₙ)*
def star(segment: Segment, *segments: Segment) -> RPQ:

    return Star(path(segment, *segments))

# (expr₁ | ... | exprₙ)
def union(segment: Segment, *segments: Segment) -> RPQ:

    pattern = _process_segment(segment)
    for seg in segments:
        pattern = Union(pattern, _process_segment(seg))

    return pattern

#-----------------------------------------------------------------------------
# UI for path finding
#-----------------------------------------------------------------------------


def match(graph, pattern, source=None, target=None,
          num_paths=None, max_length=1000,
          type="shortest"):
    """
    Find paths matching a given pattern in the graph from src to dst.

    Args:
        graph: The graph
        pattern: RPQ pattern
        src: Source node entity
        dst: Destination node entity

    Returns:
        Projected paths in the original graph
    """

    model = graph.model
    Node = graph.Node

    # Convert the pattern to an automaton
    g = pattern.glushkov()
    automaton = g.automaton()
    automaton.reduce()

    # Print the automaton for debugging
    # print(automaton)

    # Find paths via the automaton

    # allow a single source node; if so, wrap it in a relationship:
    if source is None:
        Source = Node
    elif isinstance(source, Concept) or isinstance(source, Relationship):
        Source = source
    else:
        Source = model.Relationship(f"source {{{Node}}}")
        define(Source(source))

    if target is None:
        Target = Node
    elif isinstance(target, Concept) or isinstance(target, Relationship):
        Target = target
    else:
        Target = model.Relationship(f"target {{{Node}}}")
        define(Target(target))

    # Source.inspect()
    # Target.inspect()

    product_graph, product_graph_paths = find_paths_via_automaton(
        graph, automaton, Source, Target,
        num_paths=num_paths, max_length=max_length,
        type=type)

    # Project the paths back to the original graph
    projected_paths = project_product_graph_paths_to_original_graph(product_graph, product_graph_paths, automaton)

    return projected_paths





from .metamodel import (
    Node, Capability, Reasoner, Field, Reading, Overload, Relation, Annotation,
    TypeNode, ScalarType, NumberType, UnionType, ListType, Type, Primitive, Table,
    Var, Literal, Value, Task, Logical, Sequence, Union, Match, Until, Wait, Loop, Break,
    Require, Not, Exists, Effect, Lookup, Update, Aggregate, Construct, Data, Model
)



#pyright: reportPrivateImportUsage=false
from __future__ import annotations

# Standard library imports
import re
import threading
import time
import textwrap
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, overload
import uuid


if TYPE_CHECKING:
    from rich.console import ConsoleOptions, RenderableType
    from rich.progress import Progress, TaskID

# Third-party imports
import rich
from rich.console import Console, Group
from rich.live import Live
from rich.text import Text

# Local imports
from relationalai.util.format import format_duration
from .utils import normalize_tasks


#--------------------------------------------------
# Constants
#--------------------------------------------------

# Display symbols
CHECK_MARK = "✓"
FAIL_ICON = "ⓧ"

# Spinner animation frames for running tasks
SPINNER_FRAMES = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]

# Terminal display constants
DEFAULT_TERMINAL_WIDTH = 80

# Animation constants
ANIMATION_INTERVAL = 0.1  # Seconds between animation frame updates (~10 fps)
REFRESH_RATE = 10  # Rich Live refresh rate (frames per second)
MIN_DETAIL_WIDTH = 20  # Minimum width for task details text wrapping


# ============================================================================
# Task Status Enum
# ============================================================================

class TaskStatus(str, Enum):
    """Enumeration of valid task statuses.

    Using str as the base class allows the enum values to be used as strings
    in comparisons and string operations, while providing type safety and
    preventing typos.
    """
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


# ============================================================================
# Task Class
# ============================================================================

class _RenderableWrapper:
    """Wrapper to make a callable renderable for Rich's Live.

    Rich's ``Live`` container expects a renderable object that implements
    ``__rich_console__`` and returns fresh content each time the live display
    refreshes. ``ProgressReader`` generates its layout on the fly via a
    callable, so we wrap that callable in this helper to bridge the protocol
    gap: ``Live`` sees a renderable, but internally we still recompute the
    layout lazily on every refresh.
    """

    def __init__(self, callable_fn: Callable[[], "RenderableType"]) -> None:
        """Initialize the renderable wrapper.

        Args:
            callable_fn: A callable that returns a Rich renderable object
        """
        self._callable = callable_fn

    def __rich_console__(
        self, console: Console, options: "ConsoleOptions"
    ) -> Iterator["RenderableType"]:
        """Rich protocol - yield the result of calling the callable."""
        result = self._callable()
        # Result should always be a Group, which implements __rich_console__
        yield from result.__rich_console__(console, options)

@dataclass(frozen=False)
class Task:
    """Represents a single task with its status and optional error.

    Note: Task objects are mutable only through ProgressReader APIs. When using
    ProgressReader from multiple threads, avoid modifying Task objects directly
    from different threads concurrently. Use ProgressReader helpers
    (complete_task, fail_task, update_task_status, update_task_details, etc.) for
    thread-safe updates and to ensure the UI refreshes."""

    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    text: str = ""
    details: str = ""
    status: TaskStatus = field(default=TaskStatus.RUNNING)
    error: str = ""
    _allow_direct_mutation: bool = field(default=True, init=False, repr=False, compare=False)

    _PROTECTED_FIELDS = {"status", "details", "error"}

    def __setattr__(self, name: str, value: object) -> None:
        if name in Task._PROTECTED_FIELDS:
            try:
                allow = object.__getattribute__(self, "_allow_direct_mutation")
            except AttributeError:
                allow = True
            if not allow:
                raise AttributeError(
                    "Task attributes are managed by ProgressReader. Use its public API to make changes."
                )
        object.__setattr__(self, name, value)

    def __post_init__(self):
        """Validate status values and ensure task_id."""
        # Validate that id is a non-empty string if explicitly provided
        # (If not provided, default_factory generates a UUID automatically)
        if not isinstance(self.id, str):
            raise ValueError(
                f"Task id must be a string, got {type(self.id).__name__}: {self.id!r}. "
                "Either provide a non-empty string id or omit it to auto-generate one."
            )
        if not self.id:
            raise ValueError(
                f"Task id cannot be empty. Got: {self.id!r}. "
                "Either provide a non-empty string id or omit it to auto-generate one."
            )
        # Validate status is a TaskStatus enum value
        if not isinstance(self.status, TaskStatus):
            # Allow string values for backward compatibility, but convert to enum
            try:
                self.status = TaskStatus(self.status)
            except ValueError:
                raise ValueError(
                    f"Invalid status: {self.status}. Must be one of: "
                    f"{', '.join(s.value for s in TaskStatus)}"
                )
        object.__setattr__(self, "_allow_direct_mutation", False)

    def __hash__(self) -> int:
        """Make Task hashable for use in sets.

        Uses task ID for hashing, ensuring that tasks with the same ID
        have the same hash value (required for consistent behavior in sets/dicts).
        """
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Equality comparison based on task ID.

        Two Task objects are considered equal if they have the same ID.
        This is more reliable than identity-based comparison since Task IDs
        are unique identifiers within a registry.
        """
        if not isinstance(other, Task):
            return False
        return self.id == other.id

@dataclass
class _TimingTracker:
    enabled: bool
    task_start_times: dict[Task, float] = field(default_factory=dict)
    task_end_times: dict[Task, float] = field(default_factory=dict)
    run_start: float | None = None
    run_end: float | None = None
    summary_printed: bool = False

    def track_initial_tasks(self, tasks: Iterable[Task]) -> None:
        if not self.enabled:
            return
        now = time.time()
        for task in tasks:
            self.task_start_times.setdefault(task, now)
            if task.status in (TaskStatus.COMPLETED, TaskStatus.FAILED):
                self.task_end_times.setdefault(task, now)

    def track_start(self, task: Task) -> None:
        if not self.enabled:
            return
        self.task_start_times.setdefault(task, time.time())

    def track_end(self, task: Task) -> None:
        if not self.enabled:
            return
        self.task_end_times.setdefault(task, time.time())

    def clear_end(self, task: Task) -> None:
        if not self.enabled:
            return
        self.task_end_times.pop(task, None)

    def forget(self, task: Task) -> None:
        if not self.enabled:
            return
        self.task_start_times.pop(task, None)
        self.task_end_times.pop(task, None)

    def start_run(self) -> None:
        if not self.enabled:
            return
        self.run_start = time.time()
        self.run_end = None
        self.summary_printed = False

    def stop_run(self) -> None:
        if not self.enabled:
            return
        self.run_end = time.time()

    def format_duration(self, task: Task) -> str | None:
        if not self.enabled:
            return None
        start = self.task_start_times.get(task)
        if start is None:
            return None
        end = self.task_end_times.get(task)
        if end is None:
            if task.status == TaskStatus.RUNNING:
                return None
            end = time.time()
        duration = max(0.0, end - start)
        return format_duration(duration)

    def consume_elapsed_runtime(self) -> str | None:
        if not self.enabled or self.summary_printed:
            return None
        start_candidates: list[float] = []
        if self.run_start is not None:
            start_candidates.append(self.run_start)
        start_candidates.extend(self.task_start_times.values())
        if not start_candidates:
            return None
        start_time = min(start_candidates)

        end_candidates: list[float] = []
        if self.run_end is not None:
            end_candidates.append(self.run_end)
        end_candidates.extend(self.task_end_times.values())
        if not end_candidates:
            end_candidates.append(time.time())

        elapsed = max(0.0, max(end_candidates) - start_time)
        self.summary_printed = True
        return format_duration(elapsed)

# ============================================================================
# TaskRegistry - Task Storage and Lookup
# ============================================================================

class TaskRegistry:
    """Thread-safe registry for task storage, ID lookups, and basic operations.

    This class manages the core data structure for tasks, providing:
    - Task storage and indexing
    - ID-based lookups
    - Add/remove operations
    - Completed task tracking

    All operations are thread-safe using an RLock.
    """

    def __init__(self, tasks: Iterable[Task] | None = None):
        """Initialize the task registry.

        Args:
            tasks: Optional iterable of Task objects to seed the registry with
        """
        self._lock = threading.RLock()
        self._tasks: list[Task] = []
        self._tasks_by_id: dict[str, Task] = {}
        self._completed_tasks: set[Task] = set()

        # Initialize with provided tasks
        if tasks is not None:
            for task in tasks:
                self._add_task_internal(task)

    def _add_task_internal(self, task: Task) -> None:
        """Internal method to add a task without lock (caller must hold lock)."""
        if not isinstance(task, Task):
            raise TypeError("TaskRegistry expects Task instances")

        # Assign and validate task ID
        task_id = task.id
        existing = self._tasks_by_id.get(task_id)
        if existing is not None and existing is not task:
            raise ValueError(f"Duplicate task_id detected: {task_id}")

        self._tasks.append(task)
        self._tasks_by_id[task_id] = task

    def add_task(self, task: Task | Iterable[Task]) -> None:
        """Add one or more tasks to the registry.

        Args:
            task: A Task object or iterable of Task objects to add
        """
        tasks_to_add = normalize_tasks(task)
        if not tasks_to_add:
            return

        with self._lock:
            for item in tasks_to_add:
                self._add_task_internal(item)

    def remove_task(self, task: Task | Iterable[Task]) -> bool:
        """Remove one or more tasks from the registry.

        Args:
            task: A Task object or iterable of Task objects to remove

        Returns:
            True if any tasks were removed, False if no tasks were removed.
            Note: If a task is not in the registry, it is silently ignored
            (idempotent operation).
        """
        tasks = normalize_tasks(task)
        if not tasks:
            return False

        with self._lock:
            removed = False
            for t in tasks:
                try:
                    self._tasks.remove(t)
                    self._completed_tasks.discard(t)
                    if t.id is not None:
                        self._tasks_by_id.pop(t.id, None)
                    removed = True
                except ValueError:
                    pass  # Task not found, continue
            return removed

    def get_task_by_id(self, task_id: str) -> Task | None:
        """Get a task by its ID.

        Args:
            task_id: The ID of the task to retrieve

        Returns:
            The Task object if found, None otherwise
        """
        with self._lock:
            return self._tasks_by_id.get(task_id)

    def get_task_index_by_id(self, task_id: str) -> int | None:
        """Get the index of a task by its ID.

        Args:
            task_id: The ID of the task to find

        Returns:
            The index of the task if found, None otherwise
        """
        with self._lock:
            task = self._tasks_by_id.get(task_id)
            if task is None:
                return None
            try:
                return self._tasks.index(task)
            except ValueError:
                return None

    def find_task(self, text: str) -> Task | None:
        """Find a task by its text.

        Args:
            text: The text of the task to find

        Returns:
            The Task object if found, None otherwise
        """
        with self._lock:
            for task in self._tasks:
                if task.text == text:
                    return task
        return None

    def get_task_index(self, task: Task) -> int | None:
        """Get the index of a task.

        Args:
            task: The Task object to find

        Returns:
            The index of the task if found, None otherwise
        """
        with self._lock:
            try:
                return self._tasks.index(task)
            except ValueError:
                return None

    @property
    def tasks(self) -> list[Task]:
        """Get a read-only copy of the tasks list."""
        with self._lock:
            return list(self._tasks)

    @property
    def completed_tasks(self) -> set[Task]:
        """Get a copy of the completed tasks set."""
        with self._lock:
            return set(self._completed_tasks)

    def mark_completed(self, task: Task) -> bool:
        """Mark a task as completed.

        Args:
            task: The task to mark as completed

        Returns:
            True if the task was newly marked as completed, False if already completed
        """
        with self._lock:
            if task not in self._completed_tasks:
                self._completed_tasks.add(task)
                return True
            return False

    def _unmark_completed(self, task: Task) -> None:
        """Remove a task from the completed set.

        Internal method used when a task transitions back to RUNNING status.

        Args:
            task: The task to remove from completed set
        """
        with self._lock:
            self._completed_tasks.discard(task)

    def __len__(self) -> int:
        """Return the number of tasks."""
        with self._lock:
            return len(self._tasks)

    def __contains__(self, item: object) -> bool:
        """Check if a task is in the registry."""
        with self._lock:
            return item in self._tasks


# ============================================================================
# ProgressReader - Progress Display and Task Orchestration
# ============================================================================

class ProgressReader:
    """A progress component that reads a list of tasks with progress tracking.

    Shows a progress bar and updates task status as tasks are processed.
    Can be used as a context manager or with manual start/stop methods.

    Example (context manager):
        tasks = [
            Task(text="Task 1"),
            Task(text="Task 2"),
        ]

        with ProgressReader(tasks) as reader:
            for task in reader:
                # Process task...
                reader.complete_task(task)  # Mark as done

    Example (manual control):
        reader = ProgressReader(tasks)
        reader.start()
        for task in reader:
            process_task(task)
            reader.complete_task(task)
        reader.stop()

    Note:
        For multi-threaded usage, prefer the public API methods
        (e.g., ``complete_task``, ``fail_task``, ``update_task_status``)
        instead of mutating ``Task`` attributes directly. This ensures the
        UI stays in sync and timing information remains accurate.
    """

    def __init__(
        self,
        tasks: Iterable[Task] | None = None,
        description: str = "Processing tasks",
        show_task_count: bool = True,
        show_progress_bar: bool = True,
        show_durations: bool = True,
        max_visible_tasks: int | None = 20,
        show_run_log: bool = False,
        log_file_folder: str | Path | None = "logs",
    ):
        """Initialize the ProgressReader.

        Args:
            tasks: Optional iterable of Task objects to seed the reader with
            description: Description to show in the progress bar
            show_task_count: Whether to show the "x/y" task count display (default: True)
            show_progress_bar: Whether to show the progress bar (default: True)
            show_durations: Whether to display per-task and summary durations (default: True)
            max_visible_tasks: Maximum number of active tasks (``RUNNING`` or ``FAILED``) to render at once.
                Completed tasks are omitted from the live view. Set to ``None`` to show every task.
            show_run_log: Whether to print a full task log to the console when the reader stops.
                Useful for archival/logging scenarios. Defaults to False.
            log_file_folder: Folder path where log files will be saved. A log file will be created with
                a timestamped filename (format: ``YYYYMMDD_HHMMSS.log``). Defaults to
                ``logs`` (saves in a visible folder relative to the current working directory, following
                conventions used by dbt and other major platforms). Set to ``None`` to disable file logging.
        """
        # --------------------------------------------------------------------
        # Configuration
        # --------------------------------------------------------------------
        self.description = description
        self.show_task_count = show_task_count
        self.show_progress_bar = show_progress_bar
        self.show_durations = show_durations
        self.show_run_log = show_run_log
        if max_visible_tasks is not None and max_visible_tasks <= 0:
            raise ValueError("max_visible_tasks must be a positive integer or None")
        self.max_visible_tasks = max_visible_tasks
        self.log_file_folder = Path(log_file_folder) if log_file_folder is not None else None

        # --------------------------------------------------------------------
        # Task Registry - Core data management
        # --------------------------------------------------------------------
        self._registry = TaskRegistry(tasks)

        # --------------------------------------------------------------------
        # Timing Information
        # --------------------------------------------------------------------
        self._timing = _TimingTracker(show_durations)
        if tasks is not None:
            self._timing.track_initial_tasks(tasks)

        # --------------------------------------------------------------------
        # Display Components (Rich library)
        # --------------------------------------------------------------------
        self.console = Console()
        self._progress: "Progress | None" = None
        self._task_id: "TaskID | None" = None
        self._live: Live | None = None
        self._started = False

        # --------------------------------------------------------------------
        # Animation
        # --------------------------------------------------------------------
        self._spinner_frame_index = 0  # Current spinner animation frame
        self._animation_thread: threading.Thread | None = None
        self._stop_animation = threading.Event()

        # --------------------------------------------------------------------
        # Thread Safety
        # --------------------------------------------------------------------
        self._lock = threading.RLock()  # Reentrant lock for thread-safe operations

    # ------------------------------------------------------------------------
    # Task Access (delegates to registry)
    # ------------------------------------------------------------------------

    @property
    def tasks(self) -> list[Task]:
        """Get a read-only copy of the tasks list."""
        return self._registry.tasks

    # ------------------------------------------------------------------------
    # Display Helpers
    # ------------------------------------------------------------------------

    def _get_progress_and_id(self) -> tuple["Progress", "TaskID"] | None:
        """Get progress and task_id if active, None otherwise. Helps with type narrowing."""
        if self._started and self._progress is not None and self._task_id is not None:
            return (self._progress, self._task_id)
        return None

    # ------------------------------------------------------------------------
    # Task Utilities
    # ------------------------------------------------------------------------

    def _ensure_task_list(self, task_or_tasks: Task | Iterable[Task]) -> list[Task]:
        """Convert task or iterable to list of tasks, with validation.

        This is a wrapper around the shared normalize_tasks utility for consistency.
        """
        return normalize_tasks(task_or_tasks)

    # ------------------------------------------------------------------------
    # Text Formatting
    # ------------------------------------------------------------------------

    @staticmethod
    def _wrap_detail_lines(details: str, max_width: int) -> list[str]:
        if not details:
            return []

        raw_lines = details.splitlines() or [details]
        if details.endswith("\n"):
            raw_lines.append("")

        wrapped: list[str] = []
        for raw in raw_lines:
            if raw:
                wrapped.extend(
                    textwrap.wrap(
                        raw,
                        width=max_width,
                        replace_whitespace=False,
                        drop_whitespace=False,
                    )
                    or [""]
                )
            else:
                wrapped.append("")
        return wrapped

    def _render_task_row(
        self,
        task: Task,
        spinner_char: str,
        detail_indent: str,
        max_detail_width: int,
    ) -> Text:
        line = Text("  ")
        detail_style = "dim"

        if task.status == TaskStatus.RUNNING:
            line.append(f"{spinner_char} ", style="dim")
            line.append(task.text, style="yellow dim")
        elif task.status == TaskStatus.COMPLETED:
            line.append(f"{CHECK_MARK} ")
            line.append(task.text, style="white")
            duration_text = self._timing.format_duration(task)
            if duration_text:
                line.append(f" ({duration_text})")
        elif task.status == TaskStatus.FAILED:
            error_msg = f" ({task.error})" if task.error else ""
            line = Text(f"  {FAIL_ICON} {task.text}{error_msg}", style="red")
            duration_text = self._timing.format_duration(task)
            if duration_text:
                line.append(f" ({duration_text})", style="red")
        else:
            line.append(task.text)

        for detail_line in self._wrap_detail_lines(task.details, max_detail_width):
            line.append("\n")
            line.append(detail_indent)
            if detail_line:
                line.append(detail_line, style=detail_style)

        return line

    # ------------------------------------------------------------------------
    # Task Mutation & Status Management
    # ------------------------------------------------------------------------

    def _validate_and_convert_status(self, status: TaskStatus | str) -> TaskStatus:
        """Validate and convert status to TaskStatus enum.

        Args:
            status: Status value (TaskStatus enum or string)

        Returns:
            TaskStatus enum value

        Raises:
            ValueError: If status is invalid
        """
        if isinstance(status, str):
            # Convert string to enum (for backward compatibility)
            try:
                return TaskStatus(status)
            except ValueError:
                raise ValueError(
                    f"Invalid status: {status!r}. Must be one of: "
                    f"{', '.join(s.value for s in TaskStatus)}"
                )
        elif not isinstance(status, TaskStatus):
            # Reject invalid types (e.g., int, None, etc.)
            raise ValueError(
                f"Invalid status type: {type(status).__name__}. "
                f"Expected TaskStatus enum or string, got: {status!r}"
            )
        return status

    def _update_task_attributes(
        self, task: Task, status: TaskStatus | None, error: str | None, details: str | None
    ) -> tuple[bool, bool]:
        """Update task attributes (status, error, details) and track timing.

        Args:
            task: The task to update
            status: New status to set (validated TaskStatus enum)
            error: New error message to set
            details: New details text to set

        Returns:
            Tuple of (should_refresh_totals, should_update_description)
        """
        should_refresh_totals = False
        should_update_description = False

        if status is not None:
            self._timing.track_start(task)
            object.__setattr__(task, "status", status)
            if status in (TaskStatus.COMPLETED, TaskStatus.FAILED):
                self._timing.track_end(task)
            elif status == TaskStatus.RUNNING:
                # Remove from completed set if transitioning back to RUNNING
                was_completed = task in self._registry._completed_tasks
                self._registry._unmark_completed(task)
                if was_completed:
                    should_refresh_totals = True
                self._timing.clear_end(task)
            should_update_description = True

        if error is not None:
            object.__setattr__(task, "error", error)
            should_update_description = True

        if details is not None:
            object.__setattr__(task, "details", details)
            should_update_description = True

        return should_refresh_totals, should_update_description

    def _mutate_task(
        self,
        task: Task,
        *,
        status: TaskStatus | str | None = None,
        error: str | None = None,
        details: str | None = None,
        advance: bool = True,
    ) -> bool:
        """Mutate a task's attributes (status, error, details).

        Args:
            task: The task to mutate
            status: New status to set
            error: New error message to set
            details: New details text to set
            advance: Whether to advance progress bar

        Returns:
            True if the task was found and updated, False if the task was not in registry
        """
        if status is None and error is None and details is None:
            return True  # Nothing to do, but task exists

        # Check if task exists in registry
        with self._registry._lock:
            if task not in self._registry._tasks:
                return False  # Task not in registry

            # Validate and convert status if provided
            validated_status = None
            if status is not None:
                validated_status = self._validate_and_convert_status(status)

            # Update task attributes
            should_refresh_totals, should_update_description = self._update_task_attributes(
                task, validated_status, error, details
            )

        # Handle status-specific side effects
        if validated_status == TaskStatus.COMPLETED:
            self._mark_task_completed(task, advance=advance)
        elif validated_status == TaskStatus.FAILED:
            self._mark_task_failed(task, advance=advance)

        if should_refresh_totals:
            self.update_progress_total()

        if should_update_description:
            self._update_progress(description=f"[bright_cyan]{task.text}[/bright_cyan]")

        return True  # Task was found and updated successfully

    # ------------------------------------------------------------------------
    # Progress Bar Management
    # ------------------------------------------------------------------------

    def _update_progress(self, advance: bool = False, **kwargs) -> None:
        """Update the progress bar if active.

        Args:
            advance: If True, advance the progress by 1
            **kwargs: Keyword arguments to pass to progress.update()
        """
        with self._lock:
            progress_info = self._get_progress_and_id()
            if progress_info:
                progress, task_id = progress_info
                if advance:
                    progress.advance(task_id)
                if kwargs:
                    progress.update(task_id, **kwargs)

    # ------------------------------------------------------------------------
    # Output & Logging
    # ------------------------------------------------------------------------

    def _print_elapsed_runtime(self, log_file_path: Path | None = None) -> None:
        runtime = self._timing.consume_elapsed_runtime()
        if runtime:
            rich.print()
            if log_file_path is not None:
                rich.print(f"Elapsed runtime: [cyan]{runtime}[/cyan], logs: [cyan]{log_file_path}[/cyan]")
            else:
                rich.print(f"Elapsed runtime: [cyan]{runtime}[/cyan]")

    def _print_run_log(self) -> Path | None:
        """Print run log and save to file if configured.

        Returns:
            Path to the saved log file if file logging was successful, None otherwise.
        """
        if not self.show_run_log and self.log_file_folder is None:
            return None

        with self._registry._lock:
            tasks = list(self._registry._tasks)

        if not tasks:
            return None

        detail_indent = "    "
        console_width = self.console.size.width if self.console else DEFAULT_TERMINAL_WIDTH
        max_detail_width = max(console_width - len(detail_indent) - 2, MIN_DETAIL_WIDTH)

        # Prepare log content
        log_lines: list[str] = []
        if self.show_run_log:
            rich.print()
            rich.print("[bold]Task log[/bold]")

        log_lines.append("Task log")
        log_lines.append("=" * 80)

        # Create file console once for converting Rich renderables to plain text
        # Use no_color=True and StringIO to strip ANSI escape codes from file output
        file_buffer = StringIO() if self.log_file_folder is not None else None
        file_console = Console(file=file_buffer, width=console_width, legacy_windows=False, no_color=True) if file_buffer is not None else None

        for task in tasks:
            row = self._render_task_row(task, " ", detail_indent, max_detail_width)
            if self.show_run_log:
                self.console.print(row)
            # Convert Rich renderable to plain text for file logging
            if file_console is not None and file_buffer is not None:
                file_console.print(row)
                # Get the text and strip any remaining ANSI escape codes as a safety measure
                text = file_buffer.getvalue()
                file_buffer.seek(0)
                file_buffer.truncate(0)
                # Remove ANSI escape codes (pattern matches \x1b[...m sequences)
                text = re.sub(r'\x1b\[[0-9;]*m', '', text)
                log_lines.append(text.rstrip('\n'))

        # Save to file if log_file_folder is provided
        if self.log_file_folder is not None:
            return self._save_log_to_file(log_lines)

        return None

    def _save_log_to_file(self, log_lines: list[str]) -> Path | None:
        """Save log content to a timestamped file.

        Args:
            log_lines: List of log lines to write to file.

        Returns:
            Path to the saved log file if successful, None otherwise.
        """
        if self.log_file_folder is None:
            return None

        try:
            # Create directory if it doesn't exist
            self.log_file_folder.mkdir(parents=True, exist_ok=True)

            # Generate timestamped filename
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            log_filename = f"{timestamp}.log"
            log_file_path = self.log_file_folder / log_filename

            # Write log content to file
            with open(log_file_path, "w", encoding="utf-8") as f:
                f.write("\n".join(log_lines))
                f.write("\n")

            return log_file_path
        except Exception as e:
            # Don't fail silently, but also don't break the main flow
            rich.print(f"[red]Warning: Failed to save log file: {e}[/red]")
            return None

    def _create_progress_components(self) -> tuple["Progress", "TaskID", Live]:
        """Create Rich progress components (Progress, TaskID, Live).

        Returns:
            Tuple of (progress, task_id, live) objects

        Raises:
            RuntimeError: If Live fails to start
        """
        from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn

        # Capture values needed for Rich object creation
        total = len(self._registry) if self.show_progress_bar else None
        description = self.description
        show_progress_bar = self.show_progress_bar
        show_task_count = self.show_task_count

        # Always show spinner and description
        columns = [
            SpinnerColumn(style="bright_cyan"),
            TextColumn("[bright_cyan]{task.description}[/bright_cyan]"),
        ]

        # Add progress bar components only if enabled
        if show_progress_bar:
            columns.extend([
                BarColumn(bar_width=None, complete_style="bright_cyan", finished_style="bright_cyan", pulse_style="bright_cyan"),
                TextColumn("[bright_cyan]{task.percentage:>3.0f}%"),
            ])

            # Conditionally add task count display (only relevant with progress bar)
            if show_task_count:
                columns.extend([
                    TextColumn("•"),
                    TextColumn("{task.completed}/{task.total}"),
                ])

        # Create Progress but don't start it - Live will handle rendering
        progress = Progress(*columns, console=self.console, transient=True)

        # Add the main overall progress task
        task_id = progress.add_task(description, total=total)

        # Start Live display - it will render both Progress and tasks
        # Use a wrapper class to ensure Rich recognizes it as renderable
        renderable_wrapper = _RenderableWrapper(lambda: self._render_all())
        live = Live(
            renderable_wrapper,
            console=self.console,
            refresh_per_second=REFRESH_RATE,
            vertical_overflow="crop",
        )

        return progress, task_id, live

    def start(self):
        """Start the progress display."""
        with self._lock:
            if self._started:
                return
            # Set _started early to prevent race condition with concurrent start() calls
            self._started = True

        # Create Rich objects outside lock to minimize lock duration
        progress, task_id, live = self._create_progress_components()

        self._timing.start_run()

        try:
            live.start()
        except Exception as e:
            # If Live fails to start, reset _started flag and clean up
            with self._lock:
                self._started = False
            progress.stop()
            raise RuntimeError(f"Failed to start progress display: {e}") from e

        # Update state while holding lock
        with self._lock:
            self._progress = progress
            self._task_id = task_id
            self._live = live

        # Start animation thread for spinner updates
        self._start_animation()

    def stop(self):
        """Stop the progress display."""
        # Stop animation first
        self._stop_animation_thread()

        with self._lock:
            progress_info = self._get_progress_and_id()
            if progress_info:
                progress, task_id = progress_info
                # Complete the progress bar and clear description
                progress.update(task_id, completed=len(self._registry), description="")
                # Don't call progress.stop() - Live will handle cleanup

            # Stop Live display (this will also clean up Progress)
            if self._live is not None:
                try:
                    self._live.stop()
                except Exception:
                    pass  # Ignore errors during stop
                self._live = None

            self._started = False
            self._timing.stop_run()

        # Print elapsed runtime with log file location combined
        log_file_path = self._print_run_log()
        self._print_elapsed_runtime(log_file_path)

    def __enter__(self) -> "ProgressReader":
        """Enter the context manager and start the progress display."""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context manager and stop the progress display."""
        self.stop()
        return False  # Don't suppress exceptions

    # ------------------------------------------------------------------------
    # Iterator Protocol
    # ------------------------------------------------------------------------

    def __iter__(self) -> Iterator[Task]:
        """Iterate through tasks, updating progress as we go.

        Note: Creates a snapshot of tasks and releases the lock during iteration
        to avoid deadlocks and allow concurrent modifications.
        """
        with self._registry._lock:
            tasks_copy = list(self._registry._tasks)  # Create a snapshot to iterate over

        # Lock is released here to allow concurrent modifications during iteration
        for i, task in enumerate(tasks_copy):
            # Show current task being processed
            self._update_progress(description=f"[bright_cyan]{task.text}[/bright_cyan]")

            yield task

            # Check status and update display (don't advance - iterator handles it)
            self._update_task_status_display(task, advance=False)

            # Advance overall progress (based on iteration, not completion)
            self._update_progress(advance=True)

            # If this is the last task, clear the description
            if i == len(tasks_copy) - 1:
                progress_info = self._get_progress_and_id()
                if progress_info:
                    progress, task_id = progress_info
                    progress.update(task_id, description="")

    # ------------------------------------------------------------------------
    # Rendering
    # ------------------------------------------------------------------------

    def _select_visible_tasks(self, tasks: list[Task]) -> tuple[list[Task], dict[str, int], bool]:
        """Select a subset of active tasks to render and compute hidden task statistics.

        Returns:
            Tuple of (visible_tasks, hidden_counts, all_done)
            - visible_tasks: List of tasks to display
            - hidden_counts: Dict with counts of hidden tasks by status
            - all_done: True if no RUNNING tasks remain (all completed or failed)
        """
        limit = self.max_visible_tasks

        # Single pass: filter active tasks and separate by status
        failed_tasks = []
        non_failed_tasks = []
        running_count = 0

        for task in tasks:
            if task.status == TaskStatus.COMPLETED:
                continue

            if task.status == TaskStatus.FAILED:
                failed_tasks.append(task)
            else:
                non_failed_tasks.append(task)
                if task.status == TaskStatus.RUNNING:
                    running_count += 1

        active_tasks = failed_tasks + non_failed_tasks

        if limit is None or len(active_tasks) <= limit:
            all_done = running_count == 0
            return active_tasks, {"total": 0, "RUNNING": 0, "FAILED": 0}, all_done

        # Calculate remaining slots (limit is guaranteed to be an int here due to early return above)
        remaining_slots = max(limit - len(failed_tasks), 0)

        visible = failed_tasks + non_failed_tasks[:remaining_slots]
        hidden_tasks = non_failed_tasks[remaining_slots:]

        # Count hidden tasks by status and track running tasks
        # Note: All RUNNING tasks are in non_failed_tasks (failed tasks are never RUNNING)
        hidden_running = 0
        for task in hidden_tasks:
            if task.status == TaskStatus.RUNNING:
                hidden_running += 1

        # all_done is True if no RUNNING tasks exist (neither visible nor hidden)
        all_done = running_count == 0

        counts = {
            "total": len(hidden_tasks),
            "RUNNING": hidden_running,
            "FAILED": 0,  # Failed tasks are always visible, so hidden_failed is always 0
        }

        return visible, counts, all_done

    def _render_all(self) -> Group:
        """Render both Progress and task list as a Group for Live display."""
        # Minimize lock time - only grab what we need
        with self._registry._lock:
            tasks = list(self._registry._tasks)
            spinner_char = SPINNER_FRAMES[self._spinner_frame_index]
            progress_info = self._get_progress_and_id()

        # Build task list (outside lock for better performance)
        task_lines = []

        console_width = self.console.size.width if self.console else DEFAULT_TERMINAL_WIDTH
        detail_indent = "    "
        max_detail_width = max(console_width - len(detail_indent) - 2, MIN_DETAIL_WIDTH)

        visible_tasks, hidden_counts, all_done = self._select_visible_tasks(tasks)

        # Build task lines
        for task in visible_tasks:
            task_lines.append(
                self._render_task_row(task, spinner_char, detail_indent, max_detail_width)
            )

        if hidden_counts["total"] > 0:
            hidden_parts = []
            if hidden_counts["RUNNING"]:
                hidden_parts.append(f"{hidden_counts['RUNNING']} running")
            # Note: hidden_counts["FAILED"] is always 0 because failed tasks are always visible
            detail_suffix = f" ({', '.join(hidden_parts)})" if hidden_parts else ""
            task_lines.append(
                Text(f"... {hidden_counts['total']} more active tasks hidden{detail_suffix}", style="dim")
            )

        if not visible_tasks and hidden_counts["total"] == 0 and not tasks:
            task_lines.append(Text("No active tasks", style="dim"))

        task_display = Group(*task_lines) if task_lines else Text("")

        if progress_info and not all_done:
            progress, _ = progress_info
            spacer = Text("")
            return Group(progress, spacer, task_display)

        # Either no progress bar or all tasks done
        return Group(task_display)

    # ------------------------------------------------------------------------
    # Animation
    # ------------------------------------------------------------------------

    def _start_animation(self) -> None:
        """Start the animation thread for spinner updates."""
        if self._animation_thread is not None and self._animation_thread.is_alive():
            return

        def animate():
            while not self._stop_animation.is_set():
                with self._lock:
                    self._spinner_frame_index = (self._spinner_frame_index + 1) % len(SPINNER_FRAMES)
                # Live will auto-refresh and call _render_all(), which uses the updated spinner_frame_index
                time.sleep(ANIMATION_INTERVAL)

        self._stop_animation.clear()
        self._animation_thread = threading.Thread(target=animate, daemon=True)
        self._animation_thread.start()

    def _stop_animation_thread(self) -> None:
        """Stop the animation thread."""
        self._stop_animation.set()
        if self._animation_thread is not None:
            self._animation_thread.join(timeout=0.5)

    # ------------------------------------------------------------------------
    # Task Status Handlers
    # ------------------------------------------------------------------------

    def _mark_task_completed(self, task: Task, advance: bool = True) -> None:
        """Mark a task as completed and display the result.

        Args:
            task: The task to mark as completed
            advance: Whether to advance the progress bar (True for manual completion, False during iteration)
        """
        # Mark in registry (thread-safe)
        newly_completed = self._registry.mark_completed(task)

        # Note: timing.track_end() is already called in _mutate_task, so we don't duplicate it here
        if newly_completed and advance:
            self._update_progress(advance=True)
        # Live will automatically update the display via _render_all()

    def _mark_task_failed(self, task: Task, advance: bool = True) -> None:
        """Mark a task as failed and display the result.

        Args:
            task: The task to mark as failed
            advance: Whether to advance the progress bar (True for manual completion, False during iteration)
        """
        # Mark in registry (thread-safe)
        newly_completed = self._registry.mark_completed(task)

        # Note: timing.track_end() is already called in _mutate_task, so we don't duplicate it here
        if newly_completed and advance:
            self._update_progress(advance=True)
        # Live will automatically update the display via _render_all()

    def _update_task_status_display(self, task: Task, advance: bool = True) -> None:
        """Update the display based on task status.

        Args:
            task: The task to check and display
            advance: Whether to advance the progress bar (True for manual completion, False during iteration)
        """
        if task.status == TaskStatus.COMPLETED:
            self._mark_task_completed(task, advance)
        elif task.status == TaskStatus.FAILED:
            self._mark_task_failed(task, advance)
        # RUNNING tasks are handled automatically by Live display

    # ------------------------------------------------------------------------
    # Task Lookup (delegates to registry)
    # ------------------------------------------------------------------------

    def __len__(self) -> int:
        """Return the number of tasks."""
        return len(self._registry)

    def find_task(self, text: str) -> Task | None:
        """Find a task by its text.

        Args:
            text: The text of the task to find

        Returns:
            The Task object if found, None otherwise
        """
        return self._registry.find_task(text)

    def get_task_index(self, task: Task) -> int | None:
        """Get the index of a task.

        Args:
            task: The Task object to find

        Returns:
            The index of the task if found, None otherwise
        """
        return self._registry.get_task_index(task)

    def get_task_by_id(self, task_id: str) -> Task | None:
        """Get a task by its ID.

        Args:
            task_id: The ID of the task to retrieve

        Returns:
            The Task object if found, None otherwise
        """
        return self._registry.get_task_by_id(task_id)

    def get_task_index_by_id(self, task_id: str) -> int | None:
        """Get the index of a task by its ID.

        Args:
            task_id: The ID of the task to find

        Returns:
            The index of the task if found, None otherwise
        """
        return self._registry.get_task_index_by_id(task_id)

    # ------------------------------------------------------------------------
    # Public API - Task Status Management
    # ------------------------------------------------------------------------

    @overload
    def complete_task(self, task: Task) -> bool: ...

    @overload
    def complete_task(self, task: Iterable[Task]) -> bool: ...

    def complete_task(self, task: Task | Iterable[Task]) -> bool:
        """Mark one or more tasks as completed.

        Args:
            task: A Task object or iterable of Task objects to mark as completed

        Returns:
            True if all tasks were found and updated, False if any task was not found
        """
        tasks = self._ensure_task_list(task)
        if not tasks:
            return False

        all_succeeded = True
        for t in tasks:
            success = self._mutate_task(t, status=TaskStatus.COMPLETED)
            if not success:
                all_succeeded = False

        return all_succeeded

    @overload
    def fail_task(self, task: Task, error: str = "") -> bool: ...

    @overload
    def fail_task(self, task: Iterable[Task], error: str = "") -> bool: ...

    def fail_task(self, task: Task | Iterable[Task], error: str = "") -> bool:
        """Mark one or more tasks as failed.

        Args:
            task: A Task object or iterable of Task objects to mark as failed
            error: Optional error message (applied to all tasks if iterable is provided)

        Returns:
            True if all tasks were found and updated, False if any task was not found
        """
        tasks = self._ensure_task_list(task)
        if not tasks:
            return False

        all_succeeded = True
        for t in tasks:
            success = self._mutate_task(t, status=TaskStatus.FAILED, error=error)
            if not success:
                all_succeeded = False

        return all_succeeded

    def update_task_status(self, task: Task, status: TaskStatus | str, error: str | None = None) -> bool:
        """Update the status of a task.

        Args:
            task: The task to update
            status: New status (TaskStatus enum or string for backward compatibility)
            error: Optional error message if status is "FAILED"

        Returns:
            True if the task was found and updated, False if the task was not found
        """
        return self._mutate_task(task, status=status, error=error)

    def update_task_details(self, task: Task, details: str) -> bool:
        """Update the details text of a task.

        Args:
            task: The task to update
            details: New details content to display under the task text

        Returns:
            True if the task was found and updated, False if the task was not found
        """
        return self._mutate_task(task, details=details, advance=False)

    # ------------------------------------------------------------------------
    # Public API - Task Management
    # ------------------------------------------------------------------------

    @overload
    def add_task(self, task: Task) -> None: ...

    @overload
    def add_task(self, task: Iterable[Task]) -> None: ...

    def add_task(self, task: Task | Iterable[Task]) -> None:
        """Add one or more tasks to the list and automatically update the progress bar total.

        Tasks in RUNNING status are displayed immediately when added. Tasks in other statuses
        are handled by their respective status update methods.

        Args:
            task: A Task object or iterable of Task objects to append to the reader
        """
        tasks_to_add = self._ensure_task_list(task)
        if not tasks_to_add:
            return

        # Add to registry
        self._registry.add_task(tasks_to_add)

        # Track timing for new tasks
        with self._lock:
            for item in tasks_to_add:
                self._timing.track_start(item)

            # Update totals once after all tasks are added
            self.update_progress_total()

    @overload
    def remove_task(self, task: Task) -> bool: ...

    @overload
    def remove_task(self, task: Iterable[Task]) -> bool: ...

    def remove_task(self, task: Task | Iterable[Task]) -> bool:
        """Remove one or more tasks from the list and automatically update the progress bar total.

        Args:
            task: A Task object or iterable of Task objects to remove

        Returns:
            True if any tasks were removed, False if no tasks were removed.
            Note: If a task is not in the registry, it is silently ignored
            (idempotent operation).
        """
        tasks = self._ensure_task_list(task)
        if not tasks:
            return False

        # Remove from registry
        removed = self._registry.remove_task(tasks)

        if removed:
            # Clean up timing information
            with self._lock:
                for t in tasks:
                    self._timing.forget(t)
                # Update totals after removal
                self.update_progress_total()

        return removed

    # ------------------------------------------------------------------------
    # Public API - ID-based Operations
    # ------------------------------------------------------------------------

    def complete_task_by_id(self, task_id: str) -> bool:
        """Mark the task with ``task_id`` as completed.

        Returns ``True`` if the task exists and was updated, ``False`` otherwise.
        """
        task = self.get_task_by_id(task_id)
        if task is None:
            return False
        self.complete_task(task)
        return True

    def fail_task_by_id(self, task_id: str, error: str = "") -> bool:
        """Mark the task with ``task_id`` as failed.

        Returns ``True`` if the task exists and was updated, ``False`` otherwise.
        """
        task = self.get_task_by_id(task_id)
        if task is None:
            return False
        self.fail_task(task, error=error)
        return True

    def update_task_status_by_id(self, task_id: str, status: TaskStatus | str, error: str | None = None) -> bool:
        """Update the status of the task identified by ``task_id``.

        Returns ``True`` if the task exists and was updated, ``False`` otherwise.
        """
        task = self.get_task_by_id(task_id)
        if task is None:
            return False
        self.update_task_status(task, status, error)
        return True

    def update_task_details_by_id(self, task_id: str, details: str) -> bool:
        """Update the details of the task identified by ``task_id``.

        Returns ``True`` if the task exists and was updated, ``False`` otherwise.
        """
        task = self.get_task_by_id(task_id)
        if task is None:
            return False
        self.update_task_details(task, details)
        return True

    # ------------------------------------------------------------------------
    # Progress Bar Updates
    # ------------------------------------------------------------------------

    def update_progress_total(self) -> None:
        """Update the progress bar total to match the current number of tasks.

        This is automatically called when using add_task() or remove_task().

        Also updates the completed count to preserve the current progress percentage
        when the total changes (e.g., when tasks are added dynamically).

        Note: This method may be called while already holding the lock, which is safe
        because we use RLock.
        """
        with self._lock:
            # Count how many tasks are actually completed
            completed_count = len(self._registry.completed_tasks)
            total = len(self._registry)
            # Update progress while holding lock to ensure atomicity
            progress_info = self._get_progress_and_id()
            if progress_info:
                progress, task_id = progress_info
                progress.update(task_id, total=total, completed=completed_count)



import os


def find_dbt_profiles_file() -> str | None:
    paths = [
        "./profiles.yml",
        os.path.expanduser("~/.dbt/profiles.yml"),
    ]
    for path in paths:
        if os.path.exists(path):
            return path
    return None


def find_snowflake_config_file() -> str | None:
    path = os.path.expanduser("~/.snowflake/config.toml")
    if os.path.exists(path):
        return path
    return None



import warnings

from v0.relationalai.semantics.reasoners.optimization import make_name, all_different, implies, Solver, Provider, SolverModel

__all__ = [
    "Solver",
    "Provider",
    "SolverModel",
    "make_name",
    "all_different",
    "implies",
]

warnings.warn(
    "relationalai.early_access.solvers.* is deprecated. "
    "Please migrate to relationalai.semantics.reasoners.optimization.*",
    DeprecationWarning,
    stacklevel=2,
)


from . import StringValue
from ..frontend.base import Library, Expression, Field, Value
from ..frontend.core import Float, String, Number
from decimal import Decimal as PyDecimal

# the front-end library object
library = Library("floats")

_parse_float = library.Relation("parse_float", [Field.input("value", String), Field("result", Float)])

def float(value: Value) -> Expression:
    return Float(value)

def parse_float(value: StringValue) -> Expression:
    return _parse_float(value)



"""Solver model implementation using protobuf format.

This module provides the SolverModelPB class for defining optimization and
constraint programming problems that are serialized to protobuf format and
solved by external solver engines.

Note: This protobuf-based implementation will be deprecated in favor of the
      development version (solvers_dev.py) in future releases.
"""

from __future__ import annotations

import textwrap
import time
import uuid
from typing import Any, Optional

from v0.relationalai.experimental.solvers import Solver
from v0.relationalai.semantics.internal import internal as b
from v0.relationalai.semantics.rel.executor import RelExecutor
from v0.relationalai.tools.constants import DEFAULT_QUERY_TIMEOUT_MINS
from v0.relationalai.util.timeout import calc_remaining_timeout_minutes

from .common import make_name


# =============================================================================
# Solver ProtoBuf Format Constants and Helpers
# =============================================================================

# Variable type codes for the solver protobuf format
# cont: continuous (real-valued), int: integer, bin: binary (0/1)
_VARIABLE_TYPE_CODES: dict[str, int] = {
    "cont": 40,
    "int": 41,
    "bin": 42,
}

# First-order operators: arithmetic and mathematical functions
_FIRST_ORDER_OPERATOR_CODES: dict[str, int] = {
    "+": 10,
    "-": 11,
    "*": 12,
    "/": 13,
    "^": 14,
    "abs": 20,
    "exp": 21,
    "log": 22,
    "range": 50,
}

# First-order comparison operators: relational constraints
_FIRST_ORDER_COMPARISON_CODES: dict[str, int] = {
    "=": 30,
    "!=": 31,
    "<=": 32,
    ">=": 33,
    "<": 34,
    ">": 35,
    "implies": 62,
}

# Higher-order operators: aggregation and global constraints
_HIGHER_ORDER_OPERATOR_CODES: dict[str, int] = {
    "sum": 80,
    "min": 82,
    "max": 83,
    "count": 84,
    "all_different": 90,
}

# Mapping from bound type keywords to comparison operators
_BOUND_TO_COMPARISON_OPERATOR: dict[str, str] = {
    "lower": ">=",
    "upper": "<=",
    "fixed": "=",
}


def _make_first_order_application_with_result(
    operator_code: int, *args: Any
) -> b.Expression:
    """Create a first-order application with a result variable."""
    return _make_first_order_application(operator_code, *args, b.String.ref("res"))


def _make_first_order_application(operator_code: int, *args: Any) -> b.Expression:
    """Create a first-order application expression."""
    if not (2 <= len(args) <= 4):
        raise ValueError(
            f"First-order application requires 2-4 arguments, but got {len(args)}."
        )
    result_ref = args[-1]
    if not isinstance(result_ref, b.Ref):
        raise TypeError(
            f"Last argument must be a Ref, got {type(result_ref).__name__}."
        )
    if result_ref._thing != b.String:
        result_ref = b.String.ref("res")
    application_builtin = b.Relationship.builtins["rel_primitive_solverlib_fo_appl"]
    # Wrap operands in TupleArg for the vararg solverlib_fo_appl primitive:
    # fo_appl(op, (operands...), result)
    return b.Expression(
        application_builtin, operator_code, b.TupleArg(args[:-1]), result_ref
    )


# =============================================================================
# Main Solver Model Class
# =============================================================================


class SolverModelPB:
    """Solver model interface using protobuf format for optimization problems."""

    def __init__(self, model: b.Model, num_type: str) -> None:
        """Initialize solver model.

        Args:
            model: The RelationalAI model.
            num_type: Variable type - 'cont' or 'int'.
        """
        if num_type not in ["cont", "int"]:
            raise ValueError(
                f"Invalid numerical type '{num_type}'. Must be 'cont' or 'int'."
            )
        self._model = model
        self._num_type = num_type
        self._id = next(b._global_id)
        # Maps relationships to their corresponding variable concepts
        self._variable_relationships: dict[b.Relationship, b.Concept] = {}
        prefix_uppercase = f"SolverModel_{self._id}_"
        prefix_lowercase = prefix_uppercase.lower()

        # Create core concepts for model components
        self.Variable = Variable = model.Concept(prefix_uppercase + "Variable")
        self.MinObjective = model.Concept(prefix_uppercase + "MinObjective")
        self.MaxObjective = model.Concept(prefix_uppercase + "MaxObjective")
        self.Constraint = model.Concept(prefix_uppercase + "Constraint")
        self._model_info = {
            "num_variables": Variable,
            "num_min_objectives": self.MinObjective,
            "num_max_objectives": self.MaxObjective,
            "num_constraints": self.Constraint,
        }
        # Add printed_expr property to objectives and constraints for human-readable output
        for concept in [self.MinObjective, self.MaxObjective, self.Constraint]:
            concept.printed_expr = model.Property(
                f"{{{concept._name}}} has {{printed_expr:str}}"
            )

        # Create relationships for result extraction
        result_type = "int" if num_type == "int" else "float"
        self.result_info = model.Relationship(
            "{key:str} has {value:str}", short_name=(prefix_lowercase + "result_info")
        )
        # TODO(coey) PyRel is not able to handle "Variable._name" instead of "var" below due
        # to some internal naming bug; this leads to a "Unresolved Type" warning that we
        # will have to live with for now
        self.point = model.Property(
            f"{{var}} has {{value:{result_type}}}",
            short_name=(prefix_lowercase + "point"),
        )
        self.points = model.Property(
            f"point {{i:int}} for {{var}} has {{value:{result_type}}}",
            short_name=(prefix_lowercase + "points"),
        )

        # Install raw rel to work around lack of support for rel_primitive_solverlib_print_expr
        install_rel = f"""
        declare {self.MinObjective._name}
        declare {self.MaxObjective._name}
        declare {self.Constraint._name}

        declare {prefix_lowercase}variable_name
        declare {prefix_lowercase}minobjective_name
        declare {prefix_lowercase}maxobjective_name
        declare {prefix_lowercase}constraint_name
        declare {prefix_lowercase}minobjective_serialized
        declare {prefix_lowercase}maxobjective_serialized
        declare {prefix_lowercase}constraint_serialized

        def {prefix_lowercase}minobjective_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_lowercase}minobjective_serialized[h], {prefix_lowercase}variable_name, s)

        def {prefix_lowercase}maxobjective_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_lowercase}maxobjective_serialized[h], {prefix_lowercase}variable_name, s)

        def {prefix_lowercase}constraint_printed_expr(h, s):
            rel_primitive_solverlib_print_expr({prefix_lowercase}constraint_serialized[h], {prefix_lowercase}variable_name, s)
        """
        b.define(b.RawSource("rel", textwrap.dedent(install_rel)))

    # -------------------------------------------------------------------------
    # Variable Handling
    # -------------------------------------------------------------------------

    def solve_for(
        self,
        expr,
        where: Optional[list[Any]] = None,
        populate: bool = True,
        **kwargs: Any,
    ) -> b.Concept:
        """Define decision variables.

        Args:
            expr: Relationship or expression defining variables.
            where: Optional grounding conditions.
            populate: Whether to populate relationship with solver results.
            **kwargs: Optional properties (name, type, lower, upper, fixed).

        Returns:
            Variable concept.
        """
        if where is None:
            where = []
        if isinstance(expr, b.Fragment):
            # TODO(coey): Remove in future
            raise ValueError(
                "The select fragment argument to `solve_for` is deprecated. "
                "Instead, use the `where = [conditions...]` kwarg to specify optional grounding conditions."
            )
        elif isinstance(expr, b.Expression):
            relationship = expr._op
            if not isinstance(relationship, b.Relationship):
                raise TypeError(
                    f"Expression operator must be a Relationship, got {type(relationship).__name__}."
                )
            params = expr._params
        elif isinstance(expr, b.Relationship):
            relationship = expr
            params = [
                b.field_to_type(self._model, field) for field in relationship._fields
            ]
        else:
            raise TypeError(
                f"Invalid expression type for solve_for: {type(expr).__name__}. "
                f"Expected Relationship or Expression."
            )

        if len(params) != len(relationship._fields):
            raise ValueError(
                f"Parameter count mismatch: Got {len(params)} params "
                f"but relationship has {len(relationship._fields)} fields."
            )
        if relationship in self._variable_relationships:
            raise ValueError(
                f"Variables are already defined for relationship {relationship}."
            )

        # Create a specialized Variable concept for this relationship
        # Each decision variable gets its own concept subtype
        Var = self._model.Concept(
            f"{self.Variable._name}_{str(relationship).replace('.', '_')}",
            extends=[self.Variable],
        )
        self._variable_relationships[relationship] = Var

        # Build field dict from relationship parameters (excluding the value field)
        fields = {}
        for i in range(len(params) - 1):
            if i == 0 and relationship._parent is not None:
                concept = relationship._parent
                if not isinstance(concept, b.Concept):
                    raise TypeError(
                        f"Relationship parent must be a Concept, got {type(concept).__name__}."
                    )
            else:
                concept = params[i]
            field_name = relationship._field_names[i]
            # Prevent "Implicit Subtype Relationship" warnings by explicitly registering
            # the relationship on the parent Variable concept before using it on subtypes
            self.Variable._relationships[field_name] = self.Variable._get_relationship(
                field_name
            )
            fields[field_name] = concept
        var = Var.new(**fields)
        b.define(var).where(*where)

        # Handle optional variable properties
        for key, value in kwargs.items():
            if key == "name":
                definition = self.Variable.name(var, make_name(value))
            elif key == "type":
                if not isinstance(value, str):
                    raise TypeError(
                        f"Variable 'type' must be a string, but got {type(value).__name__}."
                    )
                if value not in _VARIABLE_TYPE_CODES:
                    valid_types = ", ".join(_VARIABLE_TYPE_CODES.keys())
                    raise ValueError(
                        f"Invalid variable type '{value}'. Valid types are: {valid_types}."
                    )
                serialized_expr = _make_first_order_application_with_result(
                    _VARIABLE_TYPE_CODES[value], var
                )
                definition = self.Constraint.new(serialized=serialized_expr)
            elif key in ("lower", "upper", "fixed"):
                if not isinstance(value, (b.Producer, float, int)):
                    raise TypeError(
                        f"Variable '{key}' must be a number, but got {type(value).__name__}."
                    )
                # Map bound types to comparison operators
                operator = _BOUND_TO_COMPARISON_OPERATOR[key]
                serialized_expr = _make_first_order_application_with_result(
                    _FIRST_ORDER_COMPARISON_CODES[operator], var, value
                )
                definition = self.Constraint.new(serialized=serialized_expr)
            else:
                raise ValueError(f"Invalid keyword argument '{key}' for solve_for.")
            b.define(definition).where(*where)

        if populate:
            # Automatically populate the variable relationship with solver results
            # This defines the original relationship to pull values from self.point after solving
            value_ref = (b.Integer if self._num_type == "int" else b.Float).ref()
            b.define(
                relationship(
                    *[getattr(Var, field_name) for field_name in fields], value_ref
                )
            ).where(self.point(Var, value_ref))

        return Var

    # -------------------------------------------------------------------------
    # Objective Functions
    # -------------------------------------------------------------------------

    def minimize(
        self,
        expr: b.Producer | float | int | b.Fragment,
        name: Optional[str | list[str]] = None,
    ) -> None:
        """Add minimization objective.

        Args:
            expr: Expression to minimize.
            name: Optional objective name.
        """
        return self._add_objective(self.MinObjective, expr, name)

    def maximize(
        self,
        expr: b.Producer | float | int | b.Fragment,
        name: Optional[str | list[str]] = None,
    ) -> None:
        """Add maximization objective.

        Args:
            expr: Expression to maximize.
            name: Optional objective name.
        """
        return self._add_objective(self.MaxObjective, expr, name)

    def _add_objective(
        self,
        objective_concept: b.Concept,
        expr: b.Producer | float | int | b.Fragment,
        name: Optional[str | list[str]],
    ) -> None:
        context = SymbolifyContext(self)
        symbolic_expr = context.rewrite(expr)
        if not isinstance(symbolic_expr, Symbolic):
            # Expr is not symbolic (a constant) - wrap it as a trivial expression
            symbolic_expr = _make_first_order_application_with_result(0, expr)
        else:
            # Unwrap the symbolic expression
            unwrapped_expr = symbolic_expr.expr
            # Check if it's a bare variable (needs wrapping for protobuf)
            is_bare_variable = isinstance(unwrapped_expr, b.ConceptMember)
            is_fragment_with_variable = (
                isinstance(unwrapped_expr, b.Fragment)
                and unwrapped_expr._select
                and isinstance(unwrapped_expr._select[0], b.ConceptMember)
            )
            if is_bare_variable or is_fragment_with_variable:
                # The protobuf format requires all objectives to be expressions, not bare variables
                symbolic_expr = _make_first_order_application_with_result(
                    0, unwrapped_expr
                )
            else:
                symbolic_expr = unwrapped_expr

        if isinstance(symbolic_expr, Symbolic):
            raise ValueError(
                "Internal error. Expression is still Symbolic after unwrapping."
            )

        objective = objective_concept.new(serialized=symbolic_expr)
        definitions = [objective]
        if name is not None:
            definitions.append(objective.name(make_name(name)))
        b.define(*definitions)

    def satisfy(
        self,
        expr: b.Fragment,
        check: bool = False,
        name: Optional[str | list[str]] = None,
    ) -> None:
        """Add constraints.

        Args:
            expr: Fragment with require clause.
            check: Whether to keep require in model validation.
            name: Optional constraint name.
        """
        if not isinstance(expr, b.Fragment):
            raise TypeError(
                f"The satisfy method expects a Fragment, but got {type(expr).__name__}."
            )
        if not expr._require:
            raise ValueError("Fragment for satisfy must have a require clause.")
        if expr._select or expr._define:
            raise ValueError(
                "Fragment for satisfy must not have select or define clauses."
            )
        if not check:
            # Remove the `require` from the model roots so it is not checked as an integrity constraint
            b._remove_roots([expr])
        context = SymbolifyContext(self)

        symbolic_where_clauses = context.rewrite_where(*expr._where)
        definitions = []
        for requirement in expr._require:
            symbolic_requirement = context.rewrite(requirement)
            if not isinstance(symbolic_requirement, Symbolic):
                raise ValueError(
                    f"Cannot symbolify requirement {requirement} in satisfy. "
                    f"The requirement must contain solver variables or expressions."
                )
            constraint = self.Constraint.new(serialized=symbolic_requirement.expr)
            definitions.append(constraint)
            if name is not None:
                definitions.append(constraint.name(make_name(name)))
        b.define(*definitions).where(*symbolic_where_clauses)

    # -------------------------------------------------------------------------
    # Model Inspection and Display
    # -------------------------------------------------------------------------

    @staticmethod
    def _print_dataframe(df: Any) -> None:
        """Print dataframe with consistent formatting for long strings."""
        for row in df.itertuples(index=False):
            print("    ".join(str(val) for val in row))
        print()

    def summarize(self) -> None:
        """Print counts of variables, objectives, and constraints in the model."""
        counts_df = b.select(
            *[(b.count(item) | 0) for (_, item) in self._model_info.items()]
        ).to_df()
        if counts_df.shape != (1, 4):
            raise ValueError("Unexpected counts dataframe shape.")
        num_vars, num_min_objs, num_max_objs, num_constraints = counts_df.iloc[0]
        print(
            f"Solver model has {num_vars} variables, {num_min_objs} minimization objectives, {num_max_objs} maximization objectives, and {num_constraints} constraints."
        )

    def print(self, with_names: bool = False) -> None:
        """Print model components.

        Args:
            with_names: Whether to print expression string names (if available).
        """
        # Print variables
        var_df = b.select(self.Variable.name | "_").where(self.Variable).to_df()
        if var_df.empty:
            print("No variables defined.")
            return
        print("Solver model:")
        print()
        print(f"Variables ({var_df.shape[0]}):")
        self._print_dataframe(var_df)

        # Print components
        components = [
            (self.MinObjective, "Min objectives"),
            (self.MaxObjective, "Max objectives"),
            (self.Constraint, "Constraints"),
        ]
        printed_expr_ref = b.String.ref()
        for component_concept, component_label in components:
            selection = (
                [component_concept.name | "", printed_expr_ref]
                if with_names
                else [printed_expr_ref]
            )
            component_df = (
                b.select(*selection)
                .where(component_concept.printed_expr(printed_expr_ref))
                .to_df()
            )
            if not component_df.empty:
                print(f"{component_label} ({component_df.shape[0]}):")
                self._print_dataframe(component_df)

    # -------------------------------------------------------------------------
    # Solving and Result Handling
    # -------------------------------------------------------------------------

    def solve(
        self, solver: Solver, log_to_console: bool = False, **kwargs: Any
    ) -> None:
        """Solve the model.

        Args:
            solver: Solver instance.
            log_to_console: Whether to show solver output.
            **kwargs: Solver options and parameters.
        """
        options = {**kwargs, "version": 1}

        # Validate solver options
        for option_key, option_value in options.items():
            if not isinstance(option_key, str):
                raise TypeError(
                    f"Solver option keys must be strings, but got {type(option_key).__name__} for key {option_key!r}."
                )
            if not isinstance(option_value, (int, float, str, bool)):
                raise TypeError(
                    f"Solver option values must be int, float, str, or bool, "
                    f"but got {type(option_value).__name__} for option {option_key!r}."
                )

        # Three-phase solve process:
        # 1. Export model to Snowflake as protobuf
        # 2. Execute solver job (external solver reads from Snowflake)
        # 3. Extract and load results back into the model
        input_id = uuid.uuid4()
        model_uri = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-inputs/solver/{input_id}/model.binpb"
        sf_input_uri = f"snowflake://job-inputs/solver/{input_id}/model.binpb"
        payload: dict[str, Any] = {"solver": solver.solver_name.lower()}
        payload["options"] = options
        payload["model_uri"] = sf_input_uri

        executor = self._model._to_executor()
        if not isinstance(executor, RelExecutor):
            raise ValueError(f"Expected RelExecutor, got {type(executor).__name__}.")
        prefix_lowercase = f"solvermodel_{self._id}_"

        query_timeout_mins = kwargs.get("query_timeout_mins", None)
        config = self._model._config
        if (
            query_timeout_mins is None
            and (
                timeout_value := config.get(
                    "query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS
                )
            )
            is not None
        ):
            query_timeout_mins = int(timeout_value)
        config_file_path = getattr(config, "file_path", None)
        start_time = time.monotonic()
        remaining_timeout_minutes = query_timeout_mins

        # Step 1: Materialize the model and store it in Snowflake
        print("export model")
        # TODO(coey): Weird hack to avoid uninitialized properties error
        # This forces evaluation of the Variable concept before export
        b.select(b.count(self.Variable)).to_df()
        export_model_relation = f"""
        // TODO maybe only want to pass names if printing - like in old setup
        // Collect all model components into a relation for serialization
        def model_relation {{
            (:variable, {self.Variable._name});
            (:variable_name, {prefix_lowercase}variable_name);
            (:min_objective, {prefix_lowercase}minobjective_serialized);
            (:max_objective, {prefix_lowercase}maxobjective_serialized);
            (:constraint, {prefix_lowercase}constraint_serialized);
        }}

        @no_diagnostics(:EXPERIMENTAL)
        def model_string {{ rel_primitive_solverlib_model_string[model_relation] }}

        ic model_not_empty("Solver model is empty.") requires not empty(model_string)

        def config[:envelope, :content_type]: "application/octet-stream"
        def config[:envelope, :payload, :data]: model_string
        def config[:envelope, :payload, :path]: "{model_uri}"
        def export {{ config }}
        """
        executor.execute_raw(
            textwrap.dedent(export_model_relation),
            query_timeout_mins=remaining_timeout_minutes,
        )

        # Step 2: Execute solver job and wait for completion
        print("execute solver job")
        remaining_timeout_minutes = calc_remaining_timeout_minutes(
            start_time,
            query_timeout_mins,
            config_file_path=config_file_path,
        )
        job_id = solver._exec_job(
            payload,
            log_to_console=log_to_console,
            query_timeout_mins=remaining_timeout_minutes,
        )

        # Step 3: Extract and insert solver results into the model
        print("extract result")
        remaining_timeout_minutes = calc_remaining_timeout_minutes(
            start_time,
            query_timeout_mins,
            config_file_path=config_file_path,
        )
        extract_results_relation = f"""
        def raw_result {{
            load_binary["snowflake://APP_STATE.RAI_INTERNAL_STAGE/job-results/{job_id}/result.binpb"]
        }}

        ic result_not_empty("Solver result is empty.") requires not empty(raw_result)

        @no_diagnostics(:EXPERIMENTAL)
        def extracted {{ rel_primitive_solverlib_extract[raw_result] }}

        def delete[:{self.result_info._name}]: {self.result_info._name}
        def delete[:{self.point._name}]: {self.point._name}
        def delete[:{self.points._name}]: {self.points._name}

        def insert(:{self.result_info._name}, key, value):
            exists((original_key) | string(extracted[original_key], value) and ::std::mirror::lower(original_key, key))
        """
        if self._num_type == "int":
            insert_points_relation = f"""
            def insert(:{self.point._name}, variable, value):
                exists((float_value) | extracted(:point, variable, float_value) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], float_value, value)
                )
            def insert(:{self.points._name}, point_index, variable, value):
                exists((float_index, float_value) | extracted(:points, variable, float_index, float_value) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], float_index, point_index) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], float_value, value)
                )
            """
        else:
            insert_points_relation = f"""
            def insert(:{self.point._name}, variable, value): extracted(:point, variable, value)
            def insert(:{self.points._name}, point_index, variable, value):
                exists((float_index) | extracted(:points, variable, float_index, value) and
                    ::std::mirror::convert(std::mirror::typeof[Int128], float_index, point_index)
                )
            """
        executor.execute_raw(
            textwrap.dedent(extract_results_relation)
            + textwrap.dedent(insert_points_relation),
            readonly=False,
            query_timeout_mins=remaining_timeout_minutes,
        )

        print("finished solve")
        print()

    def load_point(self, point_index: int) -> None:
        """Load a solution point.

        Args:
            point_index: Solution point index (0-based).
        """
        if not isinstance(point_index, int):
            raise TypeError(
                f"Point index must be an integer, but got {type(point_index).__name__}."
            )
        if point_index < 0:
            raise ValueError(
                f"Point index must be non-negative, but got {point_index}."
            )
        executor = self._model._to_executor()
        if not isinstance(executor, RelExecutor):
            raise ValueError(
                f"Expected RelExecutor, but got {type(executor).__name__}."
            )
        load_point_relation = f"""
        def delete[:{self.point._name}]: {self.point._name}
        def insert(:{self.point._name}, variable, value): {self.points._name}(int128[{point_index}], variable, value)
        """
        executor.execute_raw(textwrap.dedent(load_point_relation), readonly=False)

    def summarize_result(self) -> Any:
        """Print solver result summary.

        Returns:
            DataFrame with result information.
        """
        info_keys_to_retrieve = [
            "error",
            "termination_status",
            "solve_time_sec",
            "objective_value",
            "solver_version",
            "result_count",
        ]
        key, value_ref = b.String.ref(), b.String.ref()
        result_df = (
            b.select(key, value_ref)
            .where(self.result_info(key, value_ref), key.in_(info_keys_to_retrieve))
            .to_df()
        )
        if result_df.empty:
            raise ValueError(
                "No result information is available. Has the model been solved?"
            )
        print("Solver result:")
        print(result_df.to_string(index=False, header=False))
        print()
        return result_df

    def variable_values(self, multiple: bool = False) -> b.Fragment:
        """Retrieve variable values.

        Args:
            multiple: Whether to return all solution points.

        Returns:
            Fragment for selecting values.
        """
        variable_ref = self.Variable.ref()
        value_ref = (b.Integer if self._num_type == "int" else b.Float).ref()
        if multiple:
            point_index = b.Integer.ref()
            return b.select(point_index, variable_ref.name, value_ref).where(
                self.points(point_index, variable_ref, value_ref)
            )
        return b.select(variable_ref.name, value_ref).where(
            self.point(variable_ref, value_ref)
        )

    # Valid result info keys that can be accessed as attributes
    _RESULT_INFO_KEYS = frozenset(
        [
            "error",
            "termination_status",
            "solver_version",
            "printed_model",
            "solve_time_sec",
            "objective_value",
            "result_count",
        ]
    )

    def __getattr__(self, name: str) -> Any:
        """Get result attribute (e.g., num_variables, termination_status, objective_value).

        Args:
            name: Attribute name.

        Returns:
            Attribute value or None.
        """
        # Try to get dataframe from model info or result info
        if name in self._model_info:
            result_df = b.select(b.count(self._model_info[name]) | 0).to_df()
        elif name in self._RESULT_INFO_KEYS:
            value_ref = b.String.ref()
            result_df = (
                b.select(value_ref).where(self.result_info(name, value_ref)).to_df()
            )
        else:
            return None

        # Extract and convert scalar value
        if result_df.shape != (1, 1):
            raise ValueError(
                f"Expected exactly one value for attribute '{name}', "
                f"but got dataframe with shape {result_df.shape}."
            )

        result_value = result_df.iloc[0, 0]
        if not isinstance(result_value, str):
            return result_value

        # Convert string results to appropriate types
        if name == "solve_time_sec":
            return float(result_value)
        if name == "objective_value":
            return int(result_value) if self._num_type == "int" else float(result_value)
        if name == "result_count":
            return int(result_value)
        return result_value


# =============================================================================
# Symbolic Expression Classes
# =============================================================================


class Symbolic:
    """Wrapper for symbolified solver expressions."""

    def __init__(self, expr: Any) -> None:
        if isinstance(expr, Symbolic):
            raise TypeError("Cannot wrap a Symbolic expression in another Symbolic.")
        self.expr = expr


class SymbolifyContext:
    """Context for rewriting expressions into solver-compatible symbolic form."""

    def __init__(self, solver_model: SolverModelPB) -> None:
        self.model = solver_model._model
        self.solver_model = solver_model
        # Maps original variables (or refs) to symbolic variables bound in where clauses
        self.variable_map: dict[Any, Any] = {}

    # -------------------------------------------------------------------------
    # Public Rewriting Methods
    # -------------------------------------------------------------------------

    def rewrite_where(self, *exprs: Any) -> list[Any]:
        """Rewrite where clause expressions.

        Args:
            *exprs: Where clause expressions.

        Returns:
            Rewritten expressions.
        """
        rewritten_expressions: list[Any] = []
        # Two-pass strategy: first handle variable relationships to populate variable_map,
        # then rewrite other expressions that may reference those variables
        # First pass: identify and handle variable relationship expressions
        for expression in exprs:
            if (
                isinstance(expression, b.Expression)
                and isinstance(expression._op, b.Relationship)
                and expression._op in self.solver_model._variable_relationships
            ):
                rewritten_expressions.append(
                    self._handle_variable_relationship(expression)
                )
            else:
                rewritten_expressions.append(None)
        # Second pass: rewrite remaining non-variable expressions
        for i, expr in enumerate(exprs):
            if rewritten_expressions[i] is None:
                rewritten_expressions[i] = (
                    expr
                    if expr in self.variable_map
                    else self._rewrite_nonsymbolic(expr)
                )
        return rewritten_expressions

    def rewrite(self, expr: Any) -> Optional[Symbolic | Any]:
        """Rewrite expressions to symbolify solver variables."""
        if expr is None:
            return None

        elif isinstance(expr, (int, float, str)):
            return None

        elif isinstance(expr, b.ConceptFilter):
            concept = expr._op
            assert isinstance(concept, b.Concept)
            (ident, kwargs) = expr._params
            assert ident is None
            assert isinstance(kwargs, dict)
            new_kwargs = {}
            values_were_rewritten = False
            for key, value in kwargs.items():
                rewritten_value = self.rewrite(value)
                if isinstance(rewritten_value, Symbolic):
                    raise ValueError(
                        f"Cannot symbolify ConceptFilter argument {key} with symbolic value."
                    )
                if rewritten_value is not None:
                    values_were_rewritten = True
                    new_kwargs[key] = rewritten_value
                else:
                    new_kwargs[key] = value
            if values_were_rewritten:
                return b.ConceptFilter(concept, ident, new_kwargs)
            return None

        elif isinstance(expr, (b.DataColumn, b.TypeRef, b.Concept)):
            return None

        elif isinstance(expr, b.Alias):
            return self.rewrite(expr._thing)

        elif isinstance(expr, b.Ref):
            if expr in self.variable_map:
                return Symbolic(self.variable_map[expr])
            thing = self.rewrite(expr._thing)
            if thing is not None:
                raise ValueError(
                    f"Internal error. Ref._thing rewrite unexpectedly returned {thing}."
                )
            return None

        elif isinstance(expr, b.Relationship):
            if expr in self.variable_map:
                return Symbolic(self.variable_map[expr])
            variable_result = self._get_variable_ref(expr, expr._parent)
            if variable_result is not None:
                self.variable_map[expr] = variable_result
                return Symbolic(variable_result)
            return None

        elif isinstance(expr, b.RelationshipRef):
            if expr in self.variable_map:
                return Symbolic(self.variable_map[expr])
            relationship = expr._relationship
            if isinstance(relationship, b.Relationship):
                variable_result = self._get_variable_ref(relationship, expr._parent)
                if variable_result is not None:
                    self.variable_map[expr] = variable_result
                    return Symbolic(variable_result)
            rewritten_parent = self.rewrite(expr._parent)
            if isinstance(rewritten_parent, Symbolic):
                raise ValueError(
                    "Internal error. RelationshipRef parent rewrite returned Symbolic."
                )
            if rewritten_parent is not None:
                return b.RelationshipRef(rewritten_parent, relationship)
            return None

        elif isinstance(expr, b.RelationshipFieldRef):
            relationship = expr._relationship
            if not isinstance(relationship, b.Relationship):
                # TODO(coey): Handle relationship:RelationshipReading
                return None

            # Rewrite the relationship reference
            relationship_expression = (
                relationship
                if expr._parent is None
                else b.RelationshipRef(expr._parent, relationship)
            )
            variable_result = self.rewrite(relationship_expression)
            if variable_result is None:
                return None

            # Handle symbolic result - return as-is if it's the last field
            if isinstance(variable_result, Symbolic):
                if expr._field_ix == len(relationship._fields) - 1:
                    return variable_result
                variable_result = variable_result.expr

            return getattr(variable_result, relationship._field_names[expr._field_ix])

        elif isinstance(expr, b.Expression):
            operator = self.rewrite(expr._op)
            if isinstance(operator, Symbolic):
                raise ValueError(
                    "Internal error: Expression operator rewrite returned Symbolic."
                )
            params_were_rewritten = False
            has_symbolic_params = False
            params = []
            for param in expr._params:
                rewritten_param = self.rewrite(param)
                if isinstance(rewritten_param, Symbolic):
                    has_symbolic_params = True
                    params_were_rewritten = True
                    params.append(rewritten_param.expr)
                elif rewritten_param is not None:
                    params_were_rewritten = True
                    params.append(rewritten_param)
                else:
                    params.append(param)
            if operator is not None:
                if has_symbolic_params:
                    raise NotImplementedError(
                        f"Solver rewrites cannot handle expression {expr} "
                        f"with both a symbolic operator and symbolic parameters."
                    )
                return b.Expression(operator, *params)
            if not has_symbolic_params:
                return b.Expression(expr._op, *params)
            if not params_were_rewritten:
                return None

            # Some arguments involve solver variables, so rewrite into solver protobuf format
            # This converts operations like x + y into fo_appl(ADD_OP, (x, y), res)
            if not has_symbolic_params:
                raise ValueError(
                    "Internal error. Expected symbolic parameters but none were found."
                )
            if not isinstance(expr._op, b.Relationship):
                raise NotImplementedError(
                    f"Solver rewrites cannot handle expression {expr} "
                    f"with operator type {type(expr._op).__name__}."
                )
            operator_name = expr._op._name
            if not isinstance(operator_name, str):
                raise ValueError(
                    f"Internal error. Operator name is {type(operator_name).__name__}, expected str."
                )
            if operator_name in _FIRST_ORDER_OPERATOR_CODES:
                return Symbolic(
                    _make_first_order_application(
                        _FIRST_ORDER_OPERATOR_CODES[operator_name], *params
                    )
                )
            elif operator_name in _FIRST_ORDER_COMPARISON_CODES:
                return Symbolic(
                    _make_first_order_application_with_result(
                        _FIRST_ORDER_COMPARISON_CODES[operator_name], *params
                    )
                )
            else:
                raise NotImplementedError(
                    f"Solver rewrites cannot handle operator '{operator_name}'."
                )

        elif isinstance(expr, b.Aggregate):
            # Only the last argument can be symbolic
            preceding_args = [self._rewrite_nonsymbolic(arg) for arg in expr._args[:-1]]
            group = [self._rewrite_nonsymbolic(arg) for arg in expr._group]
            # TODO(coey): Should this be done with a subcontext (for variable_map)?
            where = self.rewrite_where(*expr._where._where)
            rewritten = (
                preceding_args != expr._args[:-1]
                or group != expr._group
                or where != expr._where
            )
            symbolic_arg = self.rewrite(expr._args[-1])
            if symbolic_arg is None and not rewritten:
                return None
            if not isinstance(symbolic_arg, Symbolic):
                if symbolic_arg is None:
                    symbolic_arg = expr._args[-1]
                aggregate_expr = b.Aggregate(expr._op, *preceding_args, symbolic_arg)
                return aggregate_expr.per(*group).where(*where)

            # The last argument is symbolic - convert to higher-order application
            # Example: sum(x for x in variables) becomes ho_appl(..., x, SUM_OP)
            operator_name = expr._op._name
            if not isinstance(operator_name, str):
                raise ValueError(
                    f"Internal error. Aggregate operator name is {type(operator_name).__name__}, expected str."
                )
            if operator_name not in _HIGHER_ORDER_OPERATOR_CODES:
                raise NotImplementedError(
                    f"Solver rewrites cannot handle aggregate operator '{operator_name}'. "
                    f"Supported operators: {', '.join(_HIGHER_ORDER_OPERATOR_CODES.keys())}"
                )
            higher_order_application_builtin = b.Relationship.builtins[
                "rel_primitive_solverlib_ho_appl"
            ]
            aggregate_expr = b.Aggregate(
                higher_order_application_builtin,
                *preceding_args,
                symbolic_arg.expr,
                _HIGHER_ORDER_OPERATOR_CODES[operator_name],
            )
            return Symbolic(aggregate_expr.per(*group).where(*where))

        elif isinstance(expr, b.Union):
            # Return union of the symbolified expressions, if any are symbolic
            args_were_rewritten = False
            has_symbolic_args = False
            args = []
            for union_arg in expr._args:
                rewritten_arg = self.rewrite(union_arg)
                if isinstance(rewritten_arg, Symbolic):
                    has_symbolic_args = True
                    args.append(rewritten_arg.expr)
                elif rewritten_arg is not None:
                    args_were_rewritten = True
                    args.append(rewritten_arg)
                else:
                    args.append(union_arg)
            if has_symbolic_args:
                return Symbolic(b.union(*args))
            elif args_were_rewritten:
                return b.union(*args)
            return None

        elif isinstance(expr, b.Fragment):
            # Only support selects with one item
            if expr._define or expr._require:
                raise ValueError(
                    "Solver rewrites do not support fragments with define or require clauses."
                )
            if len(expr._select) != 1:
                raise ValueError(
                    f"Solver rewrites require fragments with exactly one select item, "
                    f"but got {len(expr._select)}."
                )
            # TODO(coey): Should this be done with a subcontext (for variable_map)?
            where = self.rewrite_where(*expr._where)
            symbolic_select = self.rewrite(expr._select[0])
            if isinstance(symbolic_select, Symbolic):
                return Symbolic(b.select(symbolic_select.expr).where(*where))
            elif symbolic_select is not None:
                return b.select(symbolic_select).where(*where)
            return None

        raise NotImplementedError(
            f"Solver rewrites cannot handle {expr} of type {type(expr).__name__}."
        )

    # -------------------------------------------------------------------------
    # Private Helper Methods
    # -------------------------------------------------------------------------

    def _handle_variable_relationship(self, expr: b.Expression) -> Any:
        """Create symbolic reference for variable relationship expression."""
        relationship = expr._op
        if not isinstance(relationship, b.Relationship):
            raise TypeError(
                f"Expected Relationship in variable expression, but got {type(relationship).__name__}."
            )
        params = expr._params
        if len(params) != len(relationship._fields):
            raise ValueError(
                f"Parameter count mismatch: Got {len(params)} params "
                f"but relationship has {len(relationship._fields)} fields."
            )
        last_param = params[-1]
        if isinstance(last_param, b.Alias):
            last_param = last_param._thing
        if not isinstance(last_param, (b.Concept, b.Ref)):
            raise TypeError(
                f"Last parameter must be a Concept or Ref, but got {type(last_param).__name__}."
            )
        # Extract and rewrite field parameters to build the symbolic variable reference
        # This maps the relationship fields to their grounding values
        fields = {}
        for i in range(len(params) - 1):
            rewritten_param = self.rewrite(params[i])
            assert not isinstance(rewritten_param, Symbolic)
            fields[relationship._field_names[i]] = (
                rewritten_param if rewritten_param is not None else params[i]
            )
        # Create new ref corresponding to the decision variable
        variable_ref = self.solver_model._variable_relationships[relationship].ref()
        self.variable_map[last_param] = variable_ref
        # Return new condition to ground the variable
        return b.where(
            *[
                getattr(variable_ref, field_name) == field_value
                for field_name, field_value in fields.items()
            ]
        )

    def _rewrite_nonsymbolic(self, expr: Any) -> Any:
        """Rewrite expression ensuring non-symbolic result."""
        new_expr = self.rewrite(expr)
        if isinstance(new_expr, Symbolic):
            raise ValueError(
                f"Internal error. Non-symbolic rewrite unexpectedly returned Symbolic for {expr}."
            )
        return expr if new_expr is None else new_expr

    def _get_variable_ref(
        self, relationship: b.Relationship, parent_producer: b.Producer | None
    ) -> Optional[Any]:
        """Get variable reference for relationship, or None if not a solver variable."""
        # Check if this relationship corresponds to a decision variable
        VariableConcept = self.solver_model._variable_relationships.get(relationship)
        if VariableConcept is None:
            return None

        properties = {}
        if parent_producer is not None:
            properties[relationship._field_names[0]] = parent_producer
        return VariableConcept(VariableConcept.ref(), **properties)



from v0.relationalai.semantics import Model, Integer, define, select, count
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.find_paths import find_shortest_paths


# First test with grid graph
model = Model("test_all", dry_run=False)

grid = Graph.construct_grid(model, 3)

path_num, k = Integer.ref(), Integer.ref()
v = grid.Node.ref()

n1_1 = grid.Node.new(row = 1, col = 1)
n1_2 = grid.Node.new(row = 1, col = 2)
n1_3 = grid.Node.new(row = 1, col = 3)
n2_1 = grid.Node.new(row = 2, col = 1)
n2_2 = grid.Node.new(row = 2, col = 2)
n2_3 = grid.Node.new(row = 2, col = 3)
n3_1 = grid.Node.new(row = 3, col = 1)
n3_2 = grid.Node.new(row = 3, col = 2)
n3_3 = grid.Node.new(row = 3, col = 3)

Source_1 = model.Concept("Source_1", extends=[grid.Node])
Target_1 = model.Concept("Target_1", extends=[grid.Node])

define(Source_1(n1_1))
define(Source_1(n1_2))
define(Target_1(n3_2))
define(Target_1(n3_3))

paths_1 = find_shortest_paths(grid, Source_1, Target_1, num_paths=1, max_length=3)
iter_path_1 = select(k, v.row, v.col).where(paths_1(path_num, k, v)).to_df()
tuple_path_1 = tuple(sorted([row for row in iter_path_1.itertuples(index = False, name = None)]))

expected_path_1 = (
    (0, 1, 2), (1, 2, 2), (2, 3, 2)
)

assert tuple_path_1 == expected_path_1


# Second test with grid graph
paths_2 = find_shortest_paths(grid, Source_1, Target_1, num_paths=10, max_length=1)

assert select(k, v.row, v.col).where(paths_2(path_num, k, v)).to_df().empty


# Third test with grid graph
Source_3 = model.Concept("Source_3", extends=[grid.Node])
Target_3 = model.Concept("Target_3", extends=[grid.Node])

define(Source_3(n1_1))
define(Target_3(n3_3))

paths_3 = find_shortest_paths(grid, Source_3, Target_3, num_paths=10, max_length=4)

assert select(count(path_num)).where(paths_3(path_num, k, v)).to_df().iloc[0,0] == 6


# Fourth test with grid graph
paths_4 = find_shortest_paths(grid, Source_3, Target_3, num_paths=10, max_length=3)

assert select(k, v.row, v.col).where(paths_4(path_num, k, v)).to_df().empty


# Fifth test with grid graph
Source_5 = model.Concept("Source_5", extends=[grid.Node])
Target_5 = model.Concept("Target_5", extends=[grid.Node])

define(Source_5(n1_1))
define(Target_5(n1_1))

paths_5 = find_shortest_paths(grid, Source_5, Target_5, num_paths=10, max_length=0)

assert select(count(path_num)).where(paths_5(path_num, k, v)).to_df().iloc[0,0] == 1


# Sixth test with grid graph
Source_6 = model.Concept("Source_6", extends=[grid.Node])
Target_6 = model.Concept("Target_6", extends=[grid.Node])

define(Source_6(n1_1))
define(Source_6(n3_3))
define(Target_6(n1_1))
define(Target_6(n3_3))

paths_6 = find_shortest_paths(grid, Source_6, Target_6, num_paths=10, max_length=0)

assert select(count(path_num)).where(paths_6(path_num, k, v)).to_df().iloc[0,0] == 2


import warnings

from v0.relationalai.semantics.metamodel.visitor import collect_by_type, Visitor, Rewriter, ReadWriteVisitor

__all__ = ["collect_by_type", "Visitor", "Rewriter", "ReadWriteVisitor"]

warnings.warn(
    "relationalai.early_access.metamodel.visitor is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.visitor",
    DeprecationWarning,
    stacklevel=2,
)


class MultipleSupertypesUnsupportedError(NotImplementedError):
    """Raised when an entity type has multiple supertypes, which is not supported."""
    def __init__(self, entity_type):
        super().__init__(f"Entity type '{entity_type}' has multiple supertypes, which is unsupported.")
        self.entity_type = entity_type


class MissingReferenceSchemeError(KeyError):
    """Raised when a reference scheme for an entity type is missing in the model."""
    def __init__(self, entity_type, lookup_chain):
        super().__init__(f"No reference scheme found for entity type '{entity_type}' in chain {lookup_chain}.")
        self.entity_type = entity_type
        self.lookup_chain = lookup_chain


from v0.relationalai.semantics.tests.test_snapshot_abstract import AbstractSnapshotTest

class SnapshotTestBase(AbstractSnapshotTest):
    """Base class for tests with snapshots."""

    def assert_match_results_snapshots(self, script_path, snapshot, handler):
        for i, msg in enumerate(handler.results, start=1):
            if "results" in msg:
                snapshot.assert_match(msg["results"].to_csv(index=False).strip('"').strip('\n"'), f"result_{i}.csv")



from __future__ import annotations
from v0.relationalai.semantics.metamodel import types, factory as f
from v0.relationalai.semantics.metamodel.util import OrderedSet
from v0.relationalai.semantics.metamodel import builtins

# Rel Annotations as IR Relations (to be used in IR Annotations)

# output in arrow
arrow = f.relation("arrow", [])
arrow_annotation = f.annotation(arrow, [])

# do not output diagnostics for this error
no_diagnostics = f.relation("no_diagnostics", [f.field("code", types.Symbol)])

# do not inline this definition
no_inline = f.relation("no_inline", [])
no_inline_annotation = f.annotation(no_inline, [])

# indicates to the rel engine that this relation is a function
function = f.relation("function", [f.input_field("code", types.Symbol)])
function_annotation = f.annotation(function, [])

inner_loop = f.relation("inner_loop", [])
inner_loop_annotation = f.annotation(inner_loop, [])

inner_loop_non_stratified = f.relation("inner_loop_non_stratified", [])
inner_loop_non_stratified_annotation = f.annotation(inner_loop_non_stratified, [])

# collect all supported builtin rel annotations
builtin_annotations = OrderedSet.from_iterable([
    arrow, no_diagnostics, no_inline, function, inner_loop, inner_loop_non_stratified,
    
    # annotations on relations that do not currently propagate into Rel
    # TODO: from Thiago, ensure annotation goes from the Logical into the proper declaration
    builtins.track,
    builtins.recursion_config,
])

builtin_annotation_names = OrderedSet.from_iterable([a.name for a in builtin_annotations])



"""
Converter for Snowflake config.toml to RAI Config format.
"""

from __future__ import annotations

from typing import Any

from .snowflake_models import SnowflakeConfigFile


def convert_snowflake_to_rai(
    sf_config: dict[str, Any],
    connection_name: str | None = None
) -> dict[str, Any]:
    config_file = SnowflakeConfigFile(**sf_config)

    # Determine connection to use (default_connection_name validity checked by model)
    if not connection_name:
        if config_file.default_connection_name:
            connection_name = config_file.default_connection_name
        elif "default" in config_file.connections:
            connection_name = "default"
        else:
            connection_name = next(iter(config_file.connections.keys()))

    if connection_name not in config_file.connections:
        available = list(config_file.connections.keys())
        raise ValueError(
            f"Connection '{connection_name}' not found. Available connections: {available}"
        )

    connection = config_file.connections[connection_name]

    # Convert connection using the model's convert() method
    connection_dict = connection.convert()

    return {
        "connections": {"snowflake": connection_dict},
        "default_connection": "snowflake",
    }
